"""Component symbols of sources."""

from manim_eng.components.base.modifiers import DiamondOuter, RoundOuter
from manim_eng.components.base.source import (
    EuropeanCurrentSourceBase,
    EuropeanVoltageSourceBase,
)

__all__ = [
    "ControlledCurrentSource",
    "ControlledVoltageSource",
    "CurrentSource",
    "VoltageSource",
]


class VoltageSource(RoundOuter, EuropeanVoltageSourceBase):
    """Circuit symbol for a voltage source."""

    def _construct(self) -> None:
        super()._construct()


class ControlledVoltageSource(DiamondOuter, EuropeanVoltageSourceBase):
    """Circuit symbol for a controlled voltage source."""

    def _construct(self) -> None:
        super()._construct()


class CurrentSource(RoundOuter, EuropeanCurrentSourceBase):
    """Circuit symbol for a current source."""

    def _construct(self) -> None:
        super()._construct()


class ControlledCurrentSource(DiamondOuter, EuropeanCurrentSourceBase):
    """Circuit symbol for a controlled current source."""

    def _construct(self) -> None:
        super()._construct()



"""manim-eng logo and implementation class."""

import manim as mn

from manim_eng.components.diodes import ZenerDiode
from manim_eng.components.resistors import Resistor
from manim_eng.components.sources import CurrentSource

__all__ = ["ManimEngLogo"]


class ManimEngLogoComponentArrangement(mn.VGroup):
    """Component arrangement for manim-eng's logo."""

    def __init__(self) -> None:
        super().__init__()

        isource = CurrentSource(
            color=mn.LOGO_GREEN, fill_color=mn.config.background_color, fill_opacity=1
        ).shift(0.5 * mn.LEFT)
        resistor = Resistor(
            color=mn.LOGO_BLUE, fill_color=mn.config.background_color, fill_opacity=1
        ).shift(0.315 * mn.UP)
        diode = (
            ZenerDiode(color=mn.LOGO_RED).rotate(90 * mn.DEGREES).shift(0.5 * mn.RIGHT)
        )

        self.add(diode, resistor, isource)


class ManimEngLogo(ManimEngLogoComponentArrangement):
    """manim-eng logo.

    Parameters
    ----------
    dark_theme : bool, optional
        Whether to adapt the logo for a dark theme (default) or a light theme.
    """

    def __init__(self, dark_theme: bool = True) -> None:
        super().__init__()

        text = mn.Tex(
            r"\textsf{manim-\textbf{eng}}",
            font_size=80,
            z_index=1,
            color=mn.LOGO_WHITE if dark_theme else mn.LOGO_BLACK,
        ).align_to(0.55 * mn.LEFT + 0.05 * mn.UP, mn.DR)

        self.add(text)
        self.shift(-self.get_center())



"""Wire implementation class."""

import abc
from typing import Any, Self

import manim as mn
from manim import typing as mnt

from manim_eng import config_eng
from manim_eng.components.base import Terminal

__all__ = ["WireBase"]


class WireBase(mn.VMobject, metaclass=abc.ABCMeta):
    """Base class for wire objects.

    Subclasses must implement the ``.get_corner_points()`` method to declare where the
    wire should have corners.
    """

    def __init__(self, start: Terminal, end: Terminal, updating: bool):
        super().__init__(stroke_width=config_eng.symbol.wire_stroke_width)

        if start == end:
            raise ValueError(
                "`start` and `end` are identical. "
                "Wires must have different terminals at each end."
            )

        self.start = start
        self.end = end

        self._attached = False

        self.__construct_wire()

        if updating:
            self.add_updater(lambda mob: mob.__construct_wire())

    def attach(self) -> Self:
        """Attach the wire to its start and end terminals, if not already attached.

        This updates the terminals so that they know they have one more connection.
        """
        if not self._attached:
            self.start._increment_connection_count()
            self.end._increment_connection_count()
            self._attached = True
        return self

    def detach(self) -> Self:
        """Detach the wire from its start and end terminals, if not already detached.

        This updates the terminals so that they know they have one fewer connection.
        """
        if self._attached:
            self.start._decrement_connection_count()
            self.end._decrement_connection_count()
            self._attached = False
        return self

    def __construct_wire(self) -> None:
        # The extra points involving the 0.001 factors extend the wire ever so slightly
        # into the terminals, producing a nice clean join between the terminals and the
        # wire
        self.set_points_as_corners(
            [
                self.start.end - 0.001 * self.start.direction,
                self.start.end,
                *self.get_corner_points(),
                self.end.end,
                self.end.end - 0.001 * self.end.direction,
            ]
        )

    @abc.abstractmethod
    def get_corner_points(self) -> list[mnt.Point3D]:
        """Get the corner points of the wire.

        Returns the vertices of the wire, not including the end points (i.e. at the
        start and end terminals).
        """

    @mn.override_animate(attach)
    def __animate_attach(self, anim_args: dict[str, Any] | None = None) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return mn.AnimationGroup(
            self.start.animate(**anim_args)._increment_connection_count(),
            self.end.animate(**anim_args)._increment_connection_count(),
        )

    @mn.override_animate(detach)
    def __animate_detach(self, anim_args: dict[str, Any] | None = None) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return mn.AnimationGroup(
            self.start.animate(**anim_args)._decrement_connection_count(),
            self.end.animate(**anim_args)._decrement_connection_count(),
        )

    @mn.override_animation(mn.Create)
    def __override_create(self, **kwargs: Any) -> mn.Animation:
        self.animate(**kwargs).attach()
        return mn.Create(self, use_override=False, **kwargs)

    @mn.override_animation(mn.Uncreate)
    def __override_uncreate(self, **kwargs: Any) -> mn.Animation:
        self.animate(**kwargs).detach()
        return mn.Uncreate(self, use_override=False, **kwargs)



"""Configuration readers for parsing TOML config files to Python dictionaries."""

import os
import tomllib
from typing import Any


class UnsupportedOsTypeError(RuntimeError):
    pass


def get_user_config() -> dict[str, Any]:
    r"""Load the user-level configuration into a dictionary.

    Loads the user-level configuration out of ``~/.config/manim/manim-eng.toml`` (on
    Linux and macOS) or ``~\AppData\Roaming\Manim\manim-eng.toml`` (on Windows) into a
    nested dictionary. The directory containing ``manim-eng.toml`` is as outlined in the
    `Manim docs <https://docs.manim.community/en/stable/guides/configuration.html#the-user-config-file>`_.

    Returns
    -------
    dict[str, Any]
        A dictionary representation of the TOML file. If the file cannot be found,
        returns an empty dictionary ``{}``.
    """
    home_directory = os.path.expanduser("~")
    config_directory: str

    match os.name:
        case "posix":
            config_directory = home_directory + "/.config/manim"
        case "nt":
            config_directory = home_directory + "/AppData/Roaming/Manim"
        case other:
            raise UnsupportedOsTypeError(f"Unsupported/unknown OS type '{other}'.")

    config_file = config_directory + "/manim-eng.toml"

    try:
        with open(config_file, "rb") as filehandle:
            return tomllib.load(filehandle)
    except OSError:
        return {}


def get_project_config() -> dict[str, Any]:
    """Load the project-level configuration from the working directory.

    Returns
    -------
    dict[str, Any]
        A dictionary representation of the TOML file. If the file cannot be found,
        returns an empty dictionary ``{}``.
    """
    config_file = os.getcwd() + "/manim-eng.toml"
    try:
        with open(config_file, "rb") as filehandle:
            return tomllib.load(filehandle)
    except OSError:
        return {}



"""Component symbols of resistor-based components."""

import manim as mn

from manim_eng._config import config_eng
from manim_eng.components.base.bipole import Bipole

__all__ = ["Resistor", "Thermistor", "VariableResistor"]

from manim_eng.components.base.modifiers import SensorModifier, VariableModifier


class Resistor(Bipole):
    """Circuit symbol for a resistor."""

    def _construct(self) -> None:
        super()._construct()
        box = mn.Rectangle(
            width=config_eng.symbol.bipole_width,
            height=config_eng.symbol.bipole_height,
        ).match_style(self)
        self._body.add(box)


class Thermistor(SensorModifier, Resistor):
    """Circuit symbol for a thermistor."""

    def _construct(self) -> None:
        super()._construct()


class VariableResistor(VariableModifier, Resistor):
    """Circuit symbol for a variable resistor."""

    def _construct(self) -> None:
        super()._construct()



"""Component symbols of monopoles, i.e. grounds and power rails."""

from typing import Any

import manim as mn
import numpy as np

from manim_eng import config_eng
from manim_eng.components.base.monopole import Monopole

__all__ = ["VDD", "VSS", "BottomRail", "Earth", "Ground", "TopRail"]


class Earth(Monopole):
    """Circuit symbol for an earthed ground connection."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(direction=mn.UP, **kwargs)

    def _construct(self) -> None:
        super()._construct()

        half_width = config_eng.symbol.monopole_width / 2
        height = half_width
        spacing = height / 3

        for line_number in range(3):
            current_half_width = half_width * (3 - line_number) / 3
            line_centre = line_number * spacing * mn.DOWN
            line = mn.Line(
                start=line_centre + current_half_width * mn.LEFT,
                end=line_centre + current_half_width * mn.RIGHT,
            ).match_style(self)
            self._body.add(line)


class Ground(Monopole):
    """Circuit symbol for a ground connection."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(direction=mn.UP, **kwargs)

    def _construct(self) -> None:
        super()._construct()

        half_width = config_eng.symbol.monopole_width / 2
        height = half_width

        symbol = mn.Polygon(
            half_width * mn.RIGHT,
            half_width * mn.LEFT,
            height * mn.DOWN,
        ).match_style(self)
        self._body.add(symbol)


class TopRail(Monopole):
    """Circuit symbol for a top rail."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(direction=mn.DOWN, **kwargs)

    def _construct(self) -> None:
        super()._construct()

        half_width = config_eng.symbol.monopole_width / 2
        line = (
            mn.Line(
                start=half_width * mn.LEFT,
                end=half_width * mn.RIGHT,
            )
            .match_style(self)
            .set_stroke(width=config_eng.symbol.wire_stroke_width)
        )
        blob = (
            mn.Dot(
                radius=config_eng.symbol.node_radius,
            )
            .match_style(self)
            .set_fill(opacity=1.0)
        )

        self._body.add(line, blob)


class BottomRail(TopRail):
    """Circuit symbol for a bottom rail."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.rotate(mn.PI, about_point=mn.ORIGIN)


class VDD(Monopole):
    """Circuit symbol for a VDD arrow."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(direction=mn.DOWN, **kwargs)

    def _construct(self) -> None:
        super()._construct()

        half_width = config_eng.symbol.monopole_width / (2 * np.sqrt(2))

        arrow = (
            mn.VMobject()
            .match_style(self)
            .set_points_as_corners(
                [
                    half_width * mn.DL,
                    mn.ORIGIN,
                    half_width * mn.DR,
                ]
            )
        )
        self._body.add(arrow)


class VSS(VDD):
    """Circuit symbol for a VSS arrow."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.rotate(mn.PI, about_point=mn.ORIGIN)



"""Component symbols of capacitor-based components."""

from typing import Any

import manim as mn

from manim_eng import config_eng
from manim_eng.components.base.bipole import Bipole
from manim_eng.components.base.modifiers import SensorModifier, VariableModifier
from manim_eng.components.base.terminal import Terminal

__all__ = ["CapacitiveSensor", "Capacitor", "VariableCapacitor"]


class Capacitor(Bipole):
    """Circuit symbol for a basic capacitor."""

    def __init__(self, **kwargs: Any) -> None:
        self.__plate_half_gap = config_eng.symbol.plate_gap / 2
        self.__plate_half_height = config_eng.symbol.plate_height / 2

        super().__init__(
            Terminal(
                position=mn.LEFT * self.__plate_half_gap,
                direction=mn.LEFT,
            ),
            Terminal(
                position=mn.RIGHT * self.__plate_half_gap,
                direction=mn.RIGHT,
            ),
            **kwargs,
        )

    def _construct(self) -> None:
        super()._construct()

        for direction in [mn.LEFT, mn.RIGHT]:
            plate_base = (
                direction * self.__plate_half_gap + mn.DOWN * self.__plate_half_height
            )
            plate = mn.Line(
                start=plate_base,
                end=plate_base + 2 * self.__plate_half_height * mn.UP,
                stroke_width=config_eng.symbol.component_stroke_width,
            ).match_style(self)
            self._body.add(plate)


class CapacitiveSensor(SensorModifier, Capacitor):
    """Circuit symbol for a capacitive sensor."""

    def _construct(self) -> None:
        super()._construct()


class VariableCapacitor(VariableModifier, Capacitor):
    """Circuit symbol for a variable capacitor."""

    def _construct(self) -> None:
        super()._construct()



"""Base classes for manim-eng components."""

from .bipole import *
from .component import *
from .modifiers import *
from .monopole import *
from .source import *
from .switch import *
from .terminal import *
from .xkcd import *



"""Component symbols of diodes."""

import manim as mn
import numpy as np

from manim_eng import config_eng
from manim_eng.components.base.bipole import SquareBipole
from manim_eng.components.base.terminal import Terminal

__all__ = ["LED", "Diode", "Photodiode", "SchottkyDiode", "TunnelDiode", "ZenerDiode"]


class Diode(SquareBipole):
    """Circuit symbol for a diode."""

    def _construct(self, draw_line: bool = True) -> None:
        """Construct a basic diode.

        Parameters
        ----------
        draw_line : bool
            Whether or not to draw the diode line. Defaults to ``True``. Set to
            ``False`` if you wish to draw the line yourself (i.e. if you're extending it
            and drawing the whole thing yourself prevents joining artefacts).
        """
        super()._construct()

        width = config_eng.symbol.square_bipole_side_length
        radius = (2 / 3) * width
        half_height = np.sqrt(3) * radius / 2
        line_start = 0.5 * width * mn.RIGHT

        triangle = (
            mn.Triangle(
                start_angle=0,
                radius=radius,
            )
            .match_style(self)
            .shift((1 / 12) * mn.LEFT)
        )
        self._body.add(triangle)

        if draw_line:
            line = (
                mn.Line(
                    start=line_start + half_height * mn.DOWN,
                    end=line_start + half_height * mn.UP,
                )
                .match_style(self)
                .set_fill(opacity=0)
            )
            self._body.add(line)

    @property
    def positive(self) -> Terminal:
        """Return the positive terminal of the diode."""
        return self.left

    @property
    def negative(self) -> Terminal:
        """Return the negative terminal of the diode."""
        return self.right

    @property
    def anode(self) -> Terminal:
        """Return the anode (positive terminal) of the diode."""
        return self.positive

    @property
    def cathode(self) -> Terminal:
        """Return the cathode (negative terminal) of the diode."""
        return self.negative


class LED(Diode):
    """Circuit symbol for an LED."""

    def _construct(self) -> None:  # type: ignore[override]
        super()._construct()

        top_left = self._body.get_corner(mn.UL)
        right = self._body.get_right()
        perp_direction = mn.normalize(right - top_left)
        perp_length = np.linalg.norm(top_left - right)
        arrow_direction = np.cross(perp_direction, mn.IN)

        arrow_length = 0.55 * perp_length

        for alpha in [0.15, 0.35]:
            arrow_start = (
                top_left + alpha * perp_direction + 0.15 * perp_length * arrow_direction
            )
            arrow_end = arrow_start + arrow_length * arrow_direction

            self._body.add(
                mn.Arrow(
                    start=arrow_start,
                    end=arrow_end,
                    buff=0,
                    tip_length=config_eng.symbol.arrow_tip_length,
                    stroke_width=config_eng.symbol.component_stroke_width,
                    color=self.stroke_color,
                    stroke_opacity=self.stroke_opacity,
                    fill_opacity=self.stroke_opacity,
                )
            )


class Photodiode(Diode):
    """Circuit symbol for a photodiode."""

    def _construct(self) -> None:  # type: ignore[override]
        super()._construct()

        top_left = self._body.get_corner(mn.UL)
        right = self._body.get_right()
        perp_direction = mn.normalize(right - top_left)
        perp_length = np.linalg.norm(top_left - right)
        arrow_direction = np.cross(perp_direction, mn.IN)

        arrow_length = 0.55 * perp_length

        for alpha in [0.15, 0.35]:
            arrow_end = (
                top_left + alpha * perp_direction + 0.15 * perp_length * arrow_direction
            )
            arrow_start = arrow_end + arrow_length * arrow_direction

            self._body.add(
                mn.Arrow(
                    start=arrow_start,
                    end=arrow_end,
                    buff=0,
                    tip_length=config_eng.symbol.arrow_tip_length,
                    stroke_width=config_eng.symbol.arrow_stroke_width,
                    color=self.stroke_color,
                    stroke_opacity=self.stroke_opacity,
                    fill_opacity=self.stroke_opacity,
                )
            )


class SchottkyDiode(Diode):
    """Circuit symbol for a Schottky diode."""

    def _construct(self) -> None:  # type: ignore[override]
        super()._construct(draw_line=False)

        top = self._body.get_corner(mn.UR)
        bottom = self._body.get_corner(mn.DR)
        height = np.linalg.norm(bottom - top)
        side_length = 0.2 * height

        line = (
            mn.VMobject()
            .match_style(self)
            .set_fill(opacity=0)
            .set_points_as_corners(
                [
                    top + side_length * mn.DR,
                    top + side_length * mn.RIGHT,
                    top,
                    bottom,
                    bottom + side_length * mn.LEFT,
                    bottom + side_length * mn.UL,
                ]
            )
        )
        self._body.add(line)


class TunnelDiode(Diode):
    """Circuit symbol for a tunnel diode."""

    def _construct(self) -> None:  # type: ignore[override]
        super()._construct(draw_line=False)

        top = self._body.get_corner(mn.UR)
        bottom = self._body.get_corner(mn.DR)
        height = np.linalg.norm(bottom - top)
        side_length = 0.2 * height

        line = (
            mn.VMobject()
            .match_style(self)
            .set_fill(opacity=0)
            .set_points_as_corners(
                [
                    top + side_length * mn.LEFT,
                    top,
                    bottom,
                    bottom + side_length * mn.LEFT,
                ]
            )
        )
        self._body.add(line)


class ZenerDiode(Diode):
    """Circuit symbol for a Zener diode."""

    def _construct(self) -> None:  # type: ignore[override]
        super()._construct(draw_line=False)

        top = self._body.get_corner(mn.UR)
        bottom = self._body.get_corner(mn.DR)
        height = np.linalg.norm(bottom - top)
        offset = 0.2 * height * mn.rotate_vector(mn.UP, angle=60 * mn.DEGREES)

        line = (
            mn.VMobject()
            .match_style(self)
            .set_fill(opacity=0)
            .set_points_as_corners(
                [
                    bottom - offset,
                    bottom,
                    top,
                    top + offset,
                ]
            )
        )
        self._body.add(line)



"""Component symbols of inductor-based components."""

import manim as mn

from manim_eng import config_eng
from manim_eng.components.base.bipole import Bipole

__all__ = ["InductiveSensor", "Inductor", "VariableInductor"]

from manim_eng.components.base.modifiers import SensorModifier, VariableModifier


class Inductor(Bipole):
    """Circuit symbol for an inductor."""

    def _construct(self) -> None:
        super()._construct()

        arc_radius = config_eng.symbol.bipole_width / 8
        for i in range(4):
            centre_x = arc_radius * (-3 + i * 2)
            arc = mn.Arc(
                radius=arc_radius,
                start_angle=mn.PI,
                angle=-mn.PI,
                arc_center=centre_x * mn.RIGHT,
                stroke_width=config_eng.symbol.component_stroke_width,
            ).match_style(self)
            self._body.add(arc)

        # Avoid the 'cut off' look at the ends of the inductor, due to the interface
        # between the terminal and inductor body
        for correction_direction in [mn.LEFT, mn.RIGHT]:
            visual_correction = (
                mn.VMobject()
                .match_style(self)
                .set_points_as_corners(
                    [
                        0.001 * mn.UP,
                        mn.ORIGIN,
                        0.001 * correction_direction,
                    ]
                )
                .shift(correction_direction * 0.5 * config_eng.symbol.bipole_width)
            )
            self._body.add(visual_correction)


class InductiveSensor(SensorModifier, Inductor):
    """Circuit symbol for an inductive sensor."""

    def _construct(self) -> None:
        super()._construct()


class VariableInductor(VariableModifier, Inductor):
    """Circuit symbol for a variable inductor."""

    def _construct(self) -> None:
        super()._construct()



"""Base class for switches and implementation helper class."""

import abc
from typing import Any, Self

import manim as mn
import numpy as np

from manim_eng import config_eng
from manim_eng.components import node
from manim_eng.components.base.bipole import Bipole
from manim_eng.components.base.terminal import Terminal

__all__ = ["BipoleSwitchBase", "PushSwitchBase"]


class BipoleSwitchBase(Bipole, metaclass=abc.ABCMeta):
    """Base class for switches with two terminals.

    Note that subclasses should construct their switch models **open**.
    """

    def __init__(self, closed: bool = False, **kwargs: Any) -> None:
        half_width = config_eng.symbol.square_bipole_side_length / 2
        self.closed: bool = closed
        self.left_node = node._create_node_blob(self, open_=True).move_to(
            half_width * mn.LEFT
        )
        self.right_node = node._create_node_blob(self, open_=True).move_to(
            half_width * mn.RIGHT
        )

        super().__init__(
            Terminal(
                position=mn.LEFT * half_width,
                direction=mn.LEFT,
            ),
            Terminal(
                position=mn.RIGHT * half_width,
                direction=mn.RIGHT,
            ),
            **kwargs,
        )

        if closed:
            self.close()

    def _construct(self) -> None:
        super()._construct()
        self._body.add(self.left_node, self.right_node)

    @abc.abstractmethod
    def open(self) -> Self:
        """Open the switch, if not already open."""

    @abc.abstractmethod
    def close(self) -> Self:
        """Close the switch, if not already closed."""

    def toggle(self) -> Self:
        """Toggle the switch position (open becomes closed, closed becomes open)."""
        if self.closed:
            return self.open()
        return self.close()

    def set_closed(self, closed: bool) -> Self:
        """Set the position of the switch."""
        if closed:
            return self.close()
        return self.open()

    @mn.override_animate(toggle)
    def __animate_toggle(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        if anim_args is None:
            anim_args = {}
        if self.closed:
            return self.animate(**anim_args).open().build()
        return self.animate(**anim_args).close().build()

    @mn.override_animate(set_closed)
    def __animate_set_closed(
        self, closed: bool, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        if anim_args is None:
            anim_args = {}
        if closed:
            return self.animate(**anim_args).close().build()
        return self.animate(**anim_args).open().build()


class PushSwitchBase(BipoleSwitchBase):
    """Circuit symbol for a basic push switch.

     The switch can be push-to-make or push-to-break.

    Parameters
    ----------
    push_to_make : bool
        Whether the switch should be push-to-make or push-to-break. ``True`` produces a
        push-to-make, ``False`` produces a push-to-break.
    closed : bool
        Whether the switch should be initially closed or not.
    """

    __travel = 1.5 * config_eng.symbol.node_radius

    def __init__(self, push_to_make: float, closed: bool, **kwargs: Any) -> None:
        self.push_to_make = push_to_make
        self.__button = mn.VGroup()
        super().__init__(closed, **kwargs)

        # Make sure the label anchor is above the greatest extension of the button
        # (i.e. when it's closed)
        if not push_to_make:
            self._label_anchor.shift(self.__travel * mn.UP)

    def _construct(self) -> None:
        super()._construct()

        if self.push_to_make:
            start = self.left_node.get_top() + self.__travel * mn.UP
            end = self.right_node.get_top() + self.__travel * mn.UP
        else:
            start = self.left_node.get_bottom() + self.__travel * mn.DOWN
            end = self.right_node.get_bottom() + self.__travel * mn.DOWN

        button_centre = self.get_top() + self.__travel * mn.UP
        button_half_width = config_eng.symbol.square_bipole_side_length / 8

        contact = mn.Line(
            start=start,
            end=end,
            stroke_width=config_eng.symbol.component_stroke_width,
        ).match_style(self)
        connector = mn.Line(
            start=contact.get_center(),
            end=button_centre,
            stroke_width=config_eng.symbol.component_stroke_width,
        ).match_style(self)
        button = mn.Line(
            start=button_centre + button_half_width * mn.LEFT,
            end=button_centre + button_half_width * mn.RIGHT,
            stroke_width=config_eng.symbol.component_stroke_width,
        ).match_style(self)
        self.__button.add(contact, connector, button)
        self._body.add(self.__button)

    def open(self) -> Self:
        """Open the switch, if not already open."""
        if not self.closed:
            return self
        direction = np.cross(
            mn.normalize(self.left_node.get_center() - self.right_node.get_center()),
            mn.OUT if self.push_to_make else mn.IN,
        )
        self.__button.shift(direction * self.__travel)
        self.closed = False
        return self

    def close(self) -> Self:
        """Close the switch, if not already closed."""
        if self.closed:
            return self
        direction = np.cross(
            mn.normalize(self.left_node.get_center() - self.right_node.get_center()),
            mn.IN if self.push_to_make else mn.OUT,
        )
        self.__button.shift(direction * self.__travel)
        self.closed = True
        return self



"""Contains the Circuit class."""

from typing import Any, Callable, Self, Sequence, cast

import manim as mn

__all__ = ["Circuit"]

from manim_eng.circuits.wire import Wire
from manim_eng.components.base.component import Component
from manim_eng.components.base.terminal import Terminal
from manim_eng.components.node import Node


class Circuit(mn.VMobject):
    """Circuit manager to conveniently handle components and their connections.

    Parameters
    ----------
    *components : Component
        Components to add to the circuit at initialisation.
    """

    def __init__(self, *components: Component) -> None:
        super().__init__()

        self.nodes = mn.VGroup()
        self.components = mn.VGroup()
        self.wires = mn.VGroup()
        super().add(self.nodes, self.components, self.wires)

        self.add(*components)

    @property
    def elements(self) -> list[Component]:
        """Returns a list of all components (including nodes) in the circuit."""
        return cast(
            list[Component], self.components.submobjects + self.nodes.submobjects
        )

    def add(self, *components: Component) -> Self:
        """Add one or more components to the circuit.

        Parameters
        ----------
        *components : Component
            The component(s) to add.
        """
        for component in components:
            # Update here to make sure that all marks are properly aligned
            component.update()
            if isinstance(component, Node):
                self.nodes.add(component)
            else:
                self.components.add(component)
        return self

    def remove(self, *components: Component) -> Self:
        """Remove one or more components from the circuit.

        Parameters
        ----------
        *components : Component
            The component(s) to remove.
        """
        self.components.remove(*components)
        return self

    def connect(self, start: Terminal, end: Terminal) -> Self:
        """Connect two terminals together.

        Parameters
        ----------
        start : Terminal
            The terminal the connecting wire should start at.
        end : Terminal
            The terminal the connecting wire should end at.

        Raises
        ------
        ValueError
            If the two terminals passed are identical.
        ValueError
            If either terminal doesn't belong to a component in this circuit.
        """
        self.__check_terminals_all_belong_to_this_circuit([start, end])
        self.wires.add(Wire(start, end).attach())
        # Nodes will potentially change their appearance on wire attachment using an
        # updater, but it needs kicking into gear
        self.nodes.update()
        return self

    def disconnect(self, *components_or_terminals: Component | Terminal) -> Self:
        """Disconnect the given components and/or terminals from one another.

        Each wire is checked to see if *both* the start *and* end terminals have been
        passed or belong to a component that was passed. If this is the case, the wire
        will be removed.

        Parameters
        ----------
        *components_or_terminals : Component | Terminal
            The group of components and terminals to disconnect from one another.

        Raises
        ------
        ValueError
            If any passed terminal does not belong to a component in this circuit.

        See Also
        --------
        isolate : Remove a wire if either of its ends is specified.
        """
        terminals = self._collapse_components_and_terminals_to_terminals(
            components_or_terminals
        )
        self.__check_terminals_all_belong_to_this_circuit(terminals)
        to_remove = self.__get_wires_from_terminal_condition(
            terminals, lambda start, end: start and end
        )
        self.wires.remove(*to_remove)
        for wire in to_remove:
            wire.detach()
        # Nodes will potentially change their appearance on wire detachment using an
        # updater, but it needs kicking into gear
        self.nodes.update()
        return self

    def isolate(self, *components_or_terminals: Component | Terminal) -> Self:
        """Remove all wires attached to each given terminal or component.

        Each wire is checked to see if either of its ends is a passed terminal or a
        terminal on a passed component. If this is the case, the wire will be removed.

        Parameters
        ----------
        *components_or_terminals : Component | Terminal
            The components and terminals to completely disconnect from the circuit.

        Raises
        ------
        ValueError
            If any passed terminal does not belong to a component in this circuit.

        See Also
        --------
        disconnect : Remove a wire if both its ends are specified.
        """
        terminals = self._collapse_components_and_terminals_to_terminals(
            components_or_terminals
        )
        self.__check_terminals_all_belong_to_this_circuit(terminals)
        to_remove = self.__get_wires_from_terminal_condition(
            terminals, lambda start, end: start or end
        )
        self.wires.remove(*to_remove)
        # Nodes will potentially change their appearance on wire detachment using an
        # updater, but it needs kicking into gear
        self.nodes.update()
        return self

    @staticmethod
    def _collapse_components_and_terminals_to_terminals(
        components_or_terminals: Sequence[Component | Terminal],
    ) -> list[Terminal]:
        terminals = []
        for component_or_terminal in components_or_terminals:
            if isinstance(component_or_terminal, Component):
                terminals.extend(component_or_terminal.terminals)
            else:
                terminals.append(component_or_terminal)
        # Remove duplicate entries
        return list(set(terminals))

    def __get_wires_from_terminal_condition(
        self, terminals: Sequence[Terminal], condition: Callable[[bool, bool], bool]
    ) -> list[Wire]:
        """Return a list of wires from the circuit based on a given condition.

        Iterates through all connections and calculates if each end of the wire is in
        ``terminals``. Whether each one is in ``terminals`` is passed to ``condition``,
        which is expected

        Parameters
        ----------
        terminals : Sequence[Terminal]
            The terminals to check all wires for.
        condition : Callable[[bool, bool], bool]
            The condition to use to determine whether a wire should be returned. Will be
            passed two booleans, whether the start or end of the wire is in
            ``terminals``, respectively, and should return ``True`` if the wire should
            be returned and ``False`` otherwise.

        Returns
        -------
        list[Wire]
            The list of wires selected by the condition.
        """
        to_remove = []
        for wire in cast(list[Wire], self.wires.submobjects):
            if condition(
                wire.start in terminals,
                wire.end in terminals,
            ):
                to_remove.append(wire)
        return to_remove

    def __check_terminals_all_belong_to_this_circuit(
        self, terminals: list[Terminal]
    ) -> None:
        terminal_set = set(terminals)
        owned_terminal_set = set()
        for component in self.elements:
            owned_terminal_set.update(component.terminals)

        terminals_not_owned = terminal_set.difference(owned_terminal_set)
        if len(terminals_not_owned) != 0:
            raise ValueError(
                f"At least one passed terminal does not "
                f"belong to any component in this circuit. "
                f"Problem terminals have the following end coordinates: "
                f"{[tuple(terminal.end) for terminal in terminals_not_owned]}"
            )

    @mn.override_animate(add)
    def __animate_add(
        self, *components: Component, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        self.add(*components)
        return mn.AnimationGroup(
            *[mn.Create(component, **anim_args) for component in components]
        )

    @mn.override_animate(remove)
    def __animate_remove(
        self, *components: Component, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        self.remove(*components)
        return mn.AnimationGroup(
            *[mn.Uncreate(component, **anim_args) for component in components]
        )

    @mn.override_animate(connect)
    def __animate_connect(
        self,
        start: Terminal,
        end: Terminal,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        self.__check_terminals_all_belong_to_this_circuit([start, end])
        new_wire = Wire(start, end)
        self.wires.add(new_wire)
        animation = mn.Create(new_wire, **anim_args)
        # This call has to be here so that the wire is properly attached when the update
        # is done
        self.nodes.update()
        return animation

    @mn.override_animate(disconnect)
    def __animate_disconnect(
        self,
        *components_or_terminals: Component | Terminal,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        terminals = self._collapse_components_and_terminals_to_terminals(
            components_or_terminals
        )
        self.__check_terminals_all_belong_to_this_circuit(terminals)
        to_remove = self.__get_wires_from_terminal_condition(
            terminals, lambda start, end: start and end
        )
        animations = [mn.Uncreate(wire, **anim_args) for wire in to_remove]
        self.wires.remove(*to_remove)

        node_update_animation = self.nodes.animate(**anim_args).update().build()
        animations.append(node_update_animation)

        return mn.AnimationGroup(*animations)

    @mn.override_animate(isolate)
    def __animate_isolate(
        self,
        *components_or_terminals: Component | Terminal,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        terminals = self._collapse_components_and_terminals_to_terminals(
            components_or_terminals
        )
        self.__check_terminals_all_belong_to_this_circuit(terminals)
        to_remove = self.__get_wires_from_terminal_condition(
            terminals, lambda start, end: start or end
        )
        animations = [mn.Uncreate(wire, **anim_args) for wire in to_remove]
        self.wires.remove(*to_remove)

        node_update_animation = self.nodes.animate(**anim_args).update().build()
        animations.append(node_update_animation)

        return mn.AnimationGroup(*animations)



"""Contains the Markable base class.

See Also
--------
mark
"""

import abc
from typing import Any, Self

import manim as mn

from manim_eng._base.mark import Mark
from manim_eng.units import Value

__all__ = ["Markable"]


class RotateMarkable(mn.Rotate):
    """Override for the Rotate animation that keeps attached marks upright."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.current_rotation = 0.0

    def interpolate_mobject(self, alpha: float) -> None:
        target_angle = self.angle * alpha
        delta_angle = target_angle - self.current_rotation
        self.mobject.rotate(delta_angle, axis=self.axis, about_point=self.about_point)
        self.current_rotation += delta_angle


class Markable(mn.VMobject, metaclass=abc.ABCMeta):
    """Base class for objects that can have marks attached.

    Parameters
    ----------
    **kwargs : Any
        Keyword arguments to pass on to ``manim.VMobject``.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.__rotate = mn.VGroup()
        self.__marks = mn.VGroup()
        super().add(self.__rotate, self.__marks)

    def rotate(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        self.__rotate.rotate(*args, **kwargs)
        self.__reposition_marks()
        return self

    def add(self, *mobjects: mn.Mobject) -> Self:
        for mobject in mobjects:
            if isinstance(mobject, Markable):
                self.__rotate.add(mobject.__rotate)
                self.__marks.add(mobject.__marks)
            else:
                self.__rotate.add(*mobjects)
        return self

    def add_to_back(self, *mobjects: mn.Mobject) -> Self:
        for mobject in mobjects:
            if isinstance(mobject, Markable):
                self.__rotate.add_to_back(mobject.__rotate)
                self.__marks.add_to_back(mobject.__marks)
            else:
                self.__rotate.add_to_back(*mobjects)
        return self

    def remove(self, *mobjects: mn.Mobject) -> Self:
        for mobject in mobjects:
            if isinstance(mobject, Markable):
                self.__rotate.remove(mobject.__rotate)
                self.__marks.remove(mobject.__marks)
            else:
                self.__rotate.remove(*mobjects)
        return self

    def _set_mark(self, mark_to_set: Mark, mark_text: str | Value) -> None:
        """Set a mark's label, adding the mark if necessary."""
        mark_text = mark_text if isinstance(mark_text, str) else mark_text.to_latex()
        self.__marks.add(mark_to_set)
        mark_to_set.set_text(mark_text)

    def _clear_mark(self, mark: Mark) -> None:
        """Clear a mark from the object."""
        self.__marks.remove(mark)

    def __reposition_marks(self) -> None:
        """Force marks to update their positions even if updating is disabled."""
        for mark in self.__marks.submobjects:
            mark._reposition()

    @mn.override_animate(_set_mark)
    def __animate_set_mark(
        self,
        mark_to_set: Mark,
        mark_text: str | Value,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        mark_text = mark_text if isinstance(mark_text, str) else mark_text.to_latex()
        if mark_to_set not in self.__marks.submobjects:
            self.__marks.add(mark_to_set)
            return mn.Create(mark_to_set.set_text(mark_text))

        mark_to_set.generate_target()
        mark_to_set.target.set_text(mark_text)
        return mn.MoveToTarget(mark_to_set, **anim_args)

    @mn.override_animate(_clear_mark)
    def __animate_clear_mark(
        self, mark_to_clear: Mark, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        anim = mn.Uncreate(mark_to_clear, remover=False, **anim_args)
        self.__marks.remove(mark_to_clear)
        return anim

    @mn.override_animation(mn.Rotate)
    def __animate_rotate(self, **kwargs: Any) -> mn.Animation:
        return RotateMarkable(self, **kwargs)



"""Contains the anchor class for attachments of marks.

See Also
--------
mark
"""

import abc

import manim as mn
import manim.typing as mnt

from manim_eng._config import config_eng

__all__ = ["Anchor"]


class Anchor(mn.Arc, metaclass=abc.ABCMeta):
    """Anchor to which Marks can be attached.

    See Also
    --------
    mark.Mark
    """

    def __init__(self, colour: mn.ManimColor) -> None:
        super().__init__(
            config_eng.anchor.radius if config_eng.debug else 0,
            start_angle=0,
            angle=2 * mn.PI,
            color=colour,
            stroke_width=config_eng.anchor.stroke_width,
            z_index=100,
        )

    @property
    def pos(self) -> mnt.Point3D:
        return self.get_center()


class AnnotationAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.annotation_colour)


class CentreAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.centre_colour)


class CurrentAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.current_colour)


class LabelAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.label_colour)


class TerminalAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.terminal_colour)


class VoltageAnchor(Anchor):
    def __init__(self) -> None:
        super().__init__(config_eng.anchor.voltage_colour)



r"""Unit definitions and surrounding Value system.

A quick note on what this module does *not* do: it does *not* set out a complete unit
system, and you will experience issues if you use it that way.

This system is designed purely for the ergonomic specification of units of a quantity at
display time, and because of this, it is designed to precisely mirror the form of the
expression as it is written in code. The below highlights a particular quirk of this.

>>> 1 / (KILO * VOLT) == 1 * KILO / VOLT
True

This is because the system stores units as they would be *written*, and you would write
:math:`\mathrm{1 / kV = 1 kV^{-1}}` and not :math:`\mathrm{1 k^{-1} V^{-1}}`.

Available prefixes
------------------

=============== ============
Prefix variable Displayed as
=============== ============
``QUETTA``      Q
``RONNA``       R
``YOTTA``       Y
``ZETTA``       Z
``EXA``         E
``PETA``        P
``TERA``        T
``GIGA``        G
``MEGA``        M
``KILO``        k
``HECTO``       h
``DECA``        da
``DECI``        d
``CENTI``       c
``MILLI``       m
``MICRO``       µ
``NANO``        n
``PICO``        p
``FEMTO``       f
``ATTO``        a
``ZEPTO``       z
``YOCTO``       y
``RONTO``       r
``QUECTO``      q
``YOBI``        Yi
``ZEBI``        Zi
``EXBI``        Ei
``PEBI``        Pi
``TEBI``        Ti
``GIBI``        Gi
``MEBI``        Mi
``KIBI``        Ki
=============== ============

Available units
---------------

Length/area/volume
^^^^^^^^^^^^^^^^^^

+---------------+--------------+
| Unit variable | Displayed as |
+===============+==============+
| ``METRE``     | m            |
+---------------+              |
| ``METER``     |              |
+---------------+--------------+
| ``ANGSTROM``  | Å            |
+---------------+--------------+
| ``MICRON``    | μm           |
+---------------+--------------+
| ``LITRE``     | L            |
+---------------+              |
| ``LITER``     |              |
+---------------+--------------+

Mass
^^^^

============= ============
Unit variable Displayed as
============= ============
``GRAM``      g
============= ============

Time/frequency
^^^^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``SECOND``    s
``MINUTE``    min
``HOUR``      hr
``HERTZ``     Hz
============= ============

Electricity
^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
AMP           A
VOLT          V
OHM           Ω
SIEMENS       S
FARAD         F
HENRY         H
COULOMB       C
============= ============

Temperature
^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``KELVIN``    K
``CELSIUS``   °C
============= ============

Illumination
^^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``CANDELA``   cd
``LUMEN``     lm
``LUX``       lx
============= ============

Quantity
^^^^^^^^
============= ============
Unit variable Displayed as
============= ============
``MOLE``      mol
============= ============


Angles
^^^^^^

============= ============
Unit variable Displayed as
============= ============
``DEGREE``    °
``RADIAN``    rad
``STERADIAN`` sr
============= ============

Energy and power
^^^^^^^^^^^^^^^^

================ ============
Unit variable    Displayed as
================ ============
``JOULE``        J
``ELECTRONVOLT`` eV
``WATT``         W
``DECIBEL``      dB
================ ============

Force and pressure
^^^^^^^^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``NEWTON``    N
``PASCAL``    Pa
``BAR``       bar
============= ============

Magnetism
^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``WEBER``     Wb
``TESLA``     T
============= ============

Digital storage
^^^^^^^^^^^^^^^

============= ============
Unit variable Displayed as
============= ============
``BIT``       b
``BYTE``      B
============= ============
"""

from __future__ import annotations

from typing import Sequence

import numpy as np

__all__ = [
    "AMP",
    "ANGSTROM",
    "ATTO",
    "BAR",
    "BIT",
    "BYTE",
    "CANDELA",
    "CELSIUS",
    "CENTI",
    "COULOMB",
    "DECA",
    "DECI",
    "DECIBEL",
    "DEGREE",
    "ELECTRONVOLT",
    "EXA",
    "EXBI",
    "FARAD",
    "FEMTO",
    "GIBI",
    "GIGA",
    "GRAM",
    "HECTO",
    "HENRY",
    "HERTZ",
    "HOUR",
    "JOULE",
    "KELVIN",
    "KIBI",
    "KILO",
    "LITER",
    "LITRE",
    "LUMEN",
    "LUX",
    "MEBI",
    "MEGA",
    "METER",
    "METRE",
    "MICRO",
    "MICRON",
    "MILLI",
    "MINUTE",
    "MOLE",
    "NANO",
    "NEWTON",
    "OHM",
    "PASCAL",
    "PEBI",
    "PETA",
    "PICO",
    "QUEBI",
    "QUECTO",
    "QUETTA",
    "RADIAN",
    "ROBI",
    "RONNA",
    "RONTO",
    "SECOND",
    "SIEMENS",
    "STERADIAN",
    "TEBI",
    "TERA",
    "TESLA",
    "VOLT",
    "WATT",
    "WEBER",
    "YOBI",
    "YOCTO",
    "YOTTA",
    "ZEBI",
    "ZEPTO",
    "ZETTA",
    "E",
    "Unit",
    "Value",
]


class Unit:
    """An SI unit or prefix.

    Parameters
    ----------
    symbol : str
        The symbol of the unit. For metres this would be 'm'.
    latex : str, optional
        The latex math mode code to present the unit symbol. If not given, defaults to
        ``symbol``.
    exponent : int | float, optional
        The exponent of the unit (e.g. 2 for :math:`m^{2}`). Defaults to 1 if
        unspecified.
    prefix : bool, optional
        Whether the symbol is that of a prefix (e.g. 'k' for 'kilo') or that of a unit.
        Defaults to ``False`` if unspecified.
    """

    def __init__(
        self,
        symbol: str,
        latex: str | None = None,
        exponent: int | float = 1,
        prefix: bool = False,
    ):
        self.symbol = symbol
        self.latex = latex if latex is not None else symbol
        self.exponent = exponent
        self.prefix = prefix

    def to_latex(self) -> str:
        """Return a LaTeX math mode string representation of the unit."""
        if self.exponent == 1:
            return self.latex
        return f"{self.latex}^{{{self.exponent}}}"

    def __mul__(self, other: Unit | UnitSequence) -> UnitSequence:
        """Create a UnitSequence from this Unit and another Unit(Sequence)."""
        if isinstance(other, Unit):
            return UnitSequence([self]) * UnitSequence([other])
        if isinstance(other, UnitSequence):
            return UnitSequence([self, *other.units])
        return NotImplemented

    def __rmul__(self, other: int | float) -> Value:
        """Create a Value with a value of other and this Unit as its unit."""
        if isinstance(other, int | float):
            return other * UnitSequence([self])
        return NotImplemented

    def __truediv__(self, other: Unit | UnitSequence) -> UnitSequence:
        """Create a UnitSequence from this Unit and one over another Unit(Sequence)."""
        if isinstance(other, Unit):
            return UnitSequence([self]) / UnitSequence([other])
        if isinstance(other, UnitSequence):
            return UnitSequence([self]) / other
        return NotImplemented

    def __rtruediv__(self, other: int | float) -> Value:
        """Create a Value with a value of other the inverse of this Unit as its unit."""
        if isinstance(other, int | float):
            return other / UnitSequence([self])
        return NotImplemented

    def __pow__(self, exponent: int | float) -> Unit:
        """Set the exponent of the unit."""
        if isinstance(exponent, int | float):
            return Unit(
                self.symbol,
                exponent=self.exponent * exponent if not self.prefix else 1,
                prefix=self.prefix,
            )
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        """Check for equality between two units."""
        if isinstance(other, Unit):
            return (
                self.symbol == other.symbol
                and self.exponent == other.exponent
                and self.prefix == other.prefix
            )
        return NotImplemented

    def __repr__(self) -> str:
        """Return a string representation of the unit."""
        if self.exponent == 1:
            return self.symbol
        return f"{self.symbol}^{self.exponent}"


class E(Unit):
    """Unit representing scientific notation. Printed as '×10^{exponent}'.

    Parameters
    ----------
    exponent: int
        The exponent of the '10'.
    """

    def __init__(self, exponent: int) -> None:
        super().__init__("×10", latex=r"\times 10", exponent=exponent)

    def to_latex(self) -> str:
        """Return a LaTeX math mode string representation of scientific notation."""
        return f"{self.latex}^{{{self.exponent}}}"

    def __repr__(self) -> str:
        """Return a string representation of scientific notation."""
        return f"{self.symbol}^{self.exponent}"


class UnitSequence:
    r"""A sequence of individual units.

    This should be considered an internal implementation type, and is not intended for
    external use.

    Parameters
    ----------
    units : Sequence[Unit]
        An (ordered) sequence of Unit that form the set of units.

    Notes
    -----
    ``units``  represent how a unit set would be *written*, so prefixes never have
    exponents. For example, :math:`1 \mathrm{V s^{-1}}` would be
    [k, :math:`\mathrm{s^{-1}}`].
    """

    def __init__(self, units: Sequence[Unit]) -> None:
        self.units = units

    def to_latex(self) -> str:
        """Return a LaTeX math mode string representation of the unit."""
        rm_internal = ""
        for unit in self.units:
            rm_internal += unit.to_latex()
            if not unit.prefix:
                rm_internal += r"\,"
        rm_internal = rm_internal.rstrip(r"\,")
        return rf"\mathrm{{{rm_internal}}}"

    def __mul__(self, other: UnitSequence | Unit) -> UnitSequence:
        """Append the RHS units to the unit sequence."""
        if isinstance(other, UnitSequence):
            return UnitSequence([*self.units, *other.units])
        if isinstance(other, Unit):
            return UnitSequence([*self.units, other])
        return NotImplemented

    def __rmul__(self, other: int | float) -> Value:
        """Create a Value with value of other and units of this sequence."""
        if isinstance(other, int | float):
            return Value(other, self)
        return NotImplemented

    def __truediv__(self, other: UnitSequence | Unit) -> UnitSequence:
        """Append the RHS units to the unit sequence with negated exponent."""
        if isinstance(other, UnitSequence):
            other_units = []
            for unit in other.units:
                other_units.append(unit**-1)
            return self * UnitSequence(other_units)
        if isinstance(other, Unit):
            return self / UnitSequence([other])
        return NotImplemented

    def __rtruediv__(self, other: int | float) -> Value:
        """Create a Value with value of other and one over the units of this sequence.

        Parameters
        ----------
        other : int | float
            The numerical value to assign to the Value instance.

        Returns
        -------
        Value
            A Value with numerical value ``other`` and the units of this instance with
            the exponents negated.
        """
        if isinstance(other, int | float):
            units = []
            for unit in self.units:
                units.append(unit**-1)
            return Value(other, UnitSequence(units))
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        """Check for equality with another Units instance."""
        if isinstance(other, UnitSequence):
            return self.units == other.units
        return NotImplemented

    def __repr__(self) -> str:
        """Return a string representation of the unit sequence."""
        to_return = ""
        for unit in self.units:
            to_return += f"{unit}"
            if not unit.prefix:
                to_return += " "
        return to_return.rstrip()


# Prefixes
QUETTA = Unit("Q", prefix=True)
RONNA = Unit("R", prefix=True)
YOTTA = Unit("Y", prefix=True)
ZETTA = Unit("Z", prefix=True)
EXA = Unit("E", prefix=True)
PETA = Unit("P", prefix=True)
TERA = Unit("T", prefix=True)
GIGA = Unit("G", prefix=True)
MEGA = Unit("M", prefix=True)
KILO = Unit("k", prefix=True)
HECTO = Unit("h", prefix=True)
DECA = Unit("da", prefix=True)
DECI = Unit("d", prefix=True)
CENTI = Unit("c", prefix=True)
MILLI = Unit("m", prefix=True)
MICRO = Unit("µ", prefix=True)
NANO = Unit("n", prefix=True)
PICO = Unit("p", prefix=True)
FEMTO = Unit("f", prefix=True)
ATTO = Unit("a", prefix=True)
ZEPTO = Unit("z", prefix=True)
YOCTO = Unit("y", prefix=True)
RONTO = Unit("r", prefix=True)
QUECTO = Unit("q", prefix=True)

QUEBI = Unit("Qi", prefix=True)
ROBI = Unit("Ri", prefix=True)
YOBI = Unit("Yi", prefix=True)
ZEBI = Unit("Zi", prefix=True)
EXBI = Unit("Ei", prefix=True)
PEBI = Unit("Pi", prefix=True)
TEBI = Unit("Ti", prefix=True)
GIBI = Unit("Gi", prefix=True)
MEBI = Unit("Mi", prefix=True)
KIBI = Unit("Ki", prefix=True)

# Length/area/volume
METRE = Unit("m")
METER = METRE
ANGSTROM = Unit("Å", latex=r"\mathring{A}")
MICRON = MICRO * METRE

LITRE = Unit("L")
LITER = LITRE

# Mass
GRAM = Unit("g")

# Time/frequency
SECOND = Unit("s")
MINUTE = Unit("min")
HOUR = Unit("hr")
HERTZ = Unit("Hz")

# Electricity
AMP = Unit("A")
VOLT = Unit("V")
OHM = Unit("Ω", r"\Omega")
SIEMENS = Unit("S")
FARAD = Unit("F")
HENRY = Unit("H")
COULOMB = Unit("C")

# Temperature
KELVIN = Unit("K")
CELSIUS = Unit("°C", latex=r"^\circ C")

# Illumination
CANDELA = Unit("cd")
LUMEN = Unit("lm")
LUX = Unit("lx")

# Quantity
MOLE = Unit("mol")

# Angles
DEGREE = Unit("°", latex=r"^\circ")
RADIAN = Unit("rad")
STERADIAN = Unit("sr")

# Energy and power
JOULE = Unit("J")
ELECTRONVOLT = Unit("eV")
WATT = Unit("W")
DECIBEL = Unit("dB")

# Force and pressure
NEWTON = Unit("N")
PASCAL = Unit("Pa")
BAR = Unit("bar")

# Magnetism
WEBER = Unit("Wb")
TESLA = Unit("T")

# Digital storage
BIT = Unit("b")
BYTE = Unit("B")

standard_engineering_prefixes = [
    QUECTO,
    RONTO,
    YOCTO,
    ZEPTO,
    ATTO,
    FEMTO,
    PICO,
    NANO,
    MICRO,
    MILLI,
    None,
    KILO,
    MEGA,
    GIGA,
    TERA,
    PETA,
    EXA,
    ZETTA,
    YOTTA,
    RONNA,
    QUETTA,
]


class Value:
    """Representation of a quantity with units for display purposes.

    This is *not* a full unit system implementation, and should not be treated as such.
    It is designed purely for being passed to display functions for them to pretty-print
    values with units with an ergonomic interface for the user.

    This constructor should not be used directly, but rather a value should be
    constructed through the product of an integer or float with a (combination of)
    units.

    >>> 2 * KILO * VOLT
    2 kV

    Parameters
    ----------
    value : int | float
        The numerical value of the quantity.
    units : UnitSequence
        The units of the quantity.
    """

    def __init__(self, value: int | float, units: UnitSequence) -> None:
        self.value = value
        self.units = units

    def to_latex(self) -> str:
        """Return a LaTeX math mode string representation of the unit."""
        spacing = r"\," if not isinstance(self.units.units[0], E) else ""
        return rf"{self.value}{spacing}{self.units.to_latex()}"

    @staticmethod
    def to_si(number: int | float) -> Value:
        """Convert a number to a factor and an SI prefix.

        Will only consider the standard multiples of 1000 (so e.g. 'c' for 'centi' will
        never be output).

        Parameters
        ----------
        number : int | float
            The value to convert.

        Returns
        -------
        Value
            A ``Value`` with the new factor and the SI prefix as its only unit.
        """
        prefix_offset = int((np.log10(float(number)) // 3))
        factor = number / 10 ** (prefix_offset * 3)
        prefix_index = (len(standard_engineering_prefixes) // 2) + prefix_offset
        prefix = standard_engineering_prefixes[prefix_index]
        units = UnitSequence([prefix] if prefix is not None else [])
        return Value(factor, units)

    def __mul__(self, other: Unit | UnitSequence) -> Value:
        """Append the RHS units to the Value's units."""
        return Value(self.value, self.units * other)

    def __truediv__(self, other: Unit | UnitSequence) -> Value:
        """Append the RHS units (with negated exponent) to the Value's units.

        Prefixes will not have their exponent negated, see the ``UnitSequence``
        documentation.
        """
        return Value(self.value, self.units / other)

    def __eq__(self, other: object) -> bool:
        """Check for equality with another Value."""
        if isinstance(other, Value):
            return self.value == other.value and self.units == other.units
        return NotImplemented

    def __repr__(self) -> str:
        """Return a string representation of the value."""
        spacing = " " if not isinstance(self.units.units[0], E) else ""
        return f"{self.value}{spacing}{self.units}"



"""Utilities for the rest of manim-eng."""

import manim as mn
import numpy as np
from manim import typing as mnt


def cardinalised(vector: mnt.Vector3D, margin: float | None = None) -> mnt.Vector3D:
    """If ``vector`` is within ``margin`` of a cardinal direction, snap it to it.

    The angle the passed ``vector`` makes with the positive horizontal is checked, and
    if it falls within ``margin`` of a given cardinal direction, i.e. up, down, left, or
    right, then the vector is snapped to that cardinal direction, maintaining its
    original magnitude.

    In the event that a vector lies perfectly on the boundary between possible snaps,
    the horizontal snap will be preferred.

    Parameters
    ----------
    vector : mnt.Vector3D
        The vector to potentially snap to a cardinal direction.
    margin : float
        The maximum angle ``vector`` can make with a cardinal direction and still be
        snapped to it, in *radians*. If not supplied, all vectors will be snapped to the
        nearest cardinal direction.

    Returns
    -------
    Vector3D
        The resultant vector.
    """
    vector_magnitude = np.linalg.norm(vector)
    angle = mn.angle_of_vector(vector)

    vector_within_margin_of_cardinal_direction = (
        (angle + margin) % (np.pi / 2) <= 2 * margin if margin is not None else True
    )
    if vector_within_margin_of_cardinal_direction:
        abs_max_index = np.argmax(np.abs(vector))
        cardinalised_vector = np.zeros_like(vector)
        # Flip the direction of the vector if necessary (i.e. if it's pointing left or
        # down)
        cardinalised_vector[abs_max_index] = vector_magnitude * np.sign(
            vector[abs_max_index]
        )
        return cardinalised_vector

    return vector



"""Contains the Component base class."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Self, cast

import manim as mn
import manim.typing as mnt
import numpy as np

from manim_eng import config_eng
from manim_eng._base.anchor import AnnotationAnchor, CentreAnchor, LabelAnchor
from manim_eng._base.mark import Mark
from manim_eng._base.markable import Markable
from manim_eng.circuits.voltage import Voltage
from manim_eng.components.base.terminal import Terminal
from manim_eng.units import Value

if TYPE_CHECKING:
    from manim_eng.components.base.monopole import Monopole
    from manim_eng.components.node import Node

__all__ = ["Component"]


class Component(Markable, metaclass=abc.ABCMeta):
    """Base class for all components.

    Parameters
    ----------
    terminals : list[Terminal]
        The terminals of the component. Management of terminal visibility is handled by
        the constructor; terminals should not be added before or after they are passed
        to this constructor.
    label : str | Value | None, optional
        A label to set. Takes a TeX math mode string, or a ``Value`` to be typeset as a
        math mode string.
    annotation : str | Value | None, optional
        An annotation to set. Takes a TeX math mode string, or a ``Value`` to be typeset
         as a math mode string.
    """

    def __init__(
        self,
        terminals: list[Terminal],
        label: str | Value | None = None,
        annotation: str | Value | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            stroke_width=config_eng.symbol.component_stroke_width, **kwargs
        )

        self._centre_anchor = CentreAnchor()
        self._label_anchor = LabelAnchor()
        self._annotation_anchor = AnnotationAnchor()

        for terminal in terminals:
            terminal.match_style(self)
        self._terminals = mn.VGroup(*terminals)
        self._body = mn.VGroup()
        self.add(self._body)

        self._construct()

        self._body.add(self._terminals)

        self.__set_up_anchors()
        self._label = Mark(self._label_anchor, self._centre_anchor)
        self._annotation = Mark(self._annotation_anchor, self._centre_anchor)
        self.__initialise_marks(label, annotation)

    def _construct(self) -> None:
        """Construct the shape of the component.

        Code to build the component's symbol goes in here  and *not* in ``__init__()``
        (contrary to Manim's standard). This is because the base ``Component`` class
        has to perform initialisation both before (to set up the groups etc.) and after
        (to set the anchor positions for annotations) the component's shape setup.

        :meta public:
        """

    @property
    def terminals(self) -> list[Terminal]:
        """The list of terminals of the component."""
        return cast(list[Terminal], self._terminals.submobjects)

    def get_center(self) -> mnt.Point3D:
        """Get the centre of the components.

        **This is not necessarily the exact centre of the box the component symbol
        occupies**. It is rather the point about which it is most logical to rotate
        the component. For bipoles, it will be at the midpoint of the line between the
        two terminals.

        Returns
        -------
        Point3D
            The centre of the components.
        """
        return self._centre_anchor.get_center()

    def align_terminal(
        self,
        self_terminal: Terminal | str,
        other: Terminal | mnt.Point3D | Node | Monopole,
        direction: mnt.Vector3D | None = None,
    ) -> Self:
        """Align a component terminal with a point or another component.

        Moves this component along the line perpendicular to ``direction`` such that the
        line between the end of ``self_terminal`` and ``other``
        has direction vector ``direction``.

        Parameters
        ----------
        self_terminal : Terminal | str
             Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a terminal (e.g. ``"right"``).
        other : Terminal | Point3D | Node | Monopole
            A ``Terminal`` belonging to another component, a ``Node``, a ``Monopole``
            (for which its single terminal is selected), or a point in space.
        direction : Vector3D | None
            The direction to align the terminals in. If not supplied, uses
            ``self_terminal``'s direction.

        Raises
        ------
        ValueError
            If a ``Terminal`` passed to ``self_terminal`` does not belong to this
            component.
        AttributeError
            If a string passed to ``self_terminal`` does not represent an existing
            attribute on this component.
        ValueError
            If a string passed to ``self_terminal`` does not represent an attribute of
            this component that produces a ``Terminal`` instance.
        ValueError
            If ``other`` belongs to this component (if it is a ``Terminal``)
            or if ``other`` *is* this component (if it is a ``Node`` or
            ``Monopole``).

        Notes
        -----
        In geometric terms, the component in moved such that the end of
        ``self_terminal`` is at the intersection of the lines that

        - Have direction vector perpendicular to ``direction`` and go through the
            current position of the end of ``self_terminal``; and
        - Have direction vector ``direction`` and go through the end of
            ``other`` (in the case that it is a ``Terminal``) or through ``other`` (in
            the case that it is a point).
        """
        from manim_eng.components.base.monopole import Monopole
        from manim_eng.components.node import Node

        self_terminal = self._get_or_check_terminal(self_terminal)
        if isinstance(other, Terminal):
            if other in self.terminals:
                raise ValueError(
                    "Terminal passed to `other_terminal` belongs to this component. "
                    "`other_terminal` should be a terminal of another component, "
                    "a point, or a separate Node or Monopole."
                )
            other = other.end
        elif isinstance(other, Node):
            if other == self:
                raise ValueError(
                    "Node passed to `other_terminal` is this component. "
                    "`other_terminal` should be a terminal of another component, "
                    "a point, or a separate Node or Monopole."
                )
            other = other.get_center()
        elif isinstance(other, Monopole):
            if other == self:
                raise ValueError(
                    "Monopole passed to `other_terminal` is this component. "
                    "`other_terminal` should be a terminal of another component, "
                    "a point, or a separate Node or Monopole."
                )
            other = other.terminal.end

        if direction is None:
            direction = self_terminal.direction

        movement_direction = np.cross(direction, mn.OUT)
        target_position = mn.find_intersection(
            [self_terminal.end],
            [movement_direction],
            [other],
            [direction],
        )[0]

        self.shift(target_position - self_terminal.end)
        return self

    def set_label(self, label: str | Value) -> Self:
        """Set the label of the component.

        Parameters
        ----------
        label : str | Value
            The label to set. Takes a TeX math mode string, or a ``Value`` to be typeset
            as a math mode string.

        See Also
        --------
        units.Value
        """
        self._set_mark(self._label, label)
        return self

    def clear_label(self) -> Self:
        """Clear the label of the component."""
        self._clear_mark(self._label)
        return self

    def set_annotation(self, annotation: str | Value) -> Self:
        """Set the annotation of the component.

        Parameters
        ----------
        annotation : str | Value
            The annotation to set. Takes a TeX math mode string, or a ``Value`` to be
            typeset as a math mode string.

        See Also
        --------
        units.Value
        """
        self._set_mark(self._annotation, annotation)
        return self

    def clear_annotation(self) -> Self:
        """Clear the annotation of the component."""
        self._clear_mark(self._annotation)
        return self

    def set_current(
        self,
        label: str | Value | None,
        terminal: Terminal | str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Set the current label on one of the terminals of the component.

        Parameters
        ----------
        label : str | Value, optional
            The current label to set. Takes a TeX math mode string, or a ``Value`` to be
            typeset as a math mode string.
        terminal : Terminal | str, optional
            Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a component (e.g. ``"right"``).
            If unspecified, defaults to the first terminal in the internal terminal
            list. In the case of monopoles, this is the only terminal, and in the case
            of bipoles, this is the left terminal.

        Other Parameters
        ----------------
        **kwargs : Any
            Kwargs to pass on to the terminal's ``.set_current()`` method.

        Raises
        ------
        ValueError
            If a ``Terminal`` passed to ``terminal`` does not belong to this component.
        AttributeError
            If a string passed to ``terminal`` does not represent an existing attribute.
        ValueError
            If a string passed to ``terminal`` does not represent an attribute of this
            component that produces a ``Terminal`` instance.

        See Also
        --------
        components.base.terminal.Terminal.set_current()
        """
        terminal = self._get_or_check_terminal(terminal)
        terminal.set_current(label=label, **kwargs)
        return self

    def reset_current(
        self, label: str | Value, terminal: Terminal | str | None = None, **kwargs: Any
    ) -> Self:
        """Reset the current label on one of the terminals of the component.

        **Warning:** Using this will reset all unspecified arguments to their default
        values. See the documentation of ``Terminal.reset_current()`` for more
        information.

        Parameters
        ----------
        label : str | Value
            The current label to set. Takes a TeX math mode string, or a ``Value`` to be
            typeset as a math mode string.
        terminal : Terminal | str | None
            Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a component (e.g. ``"right"``).
            If unspecified, defaults to the first terminal in the internal terminal
            list. In the case of monopoles, this is the only terminal, and in the case
            of bipoles, this is the left terminal.

        Other Parameters
        ----------------
        **kwargs : Any
            Kwargs to pass on to the terminal's ``.set_current()`` method.

        Raises
        ------
        ValueError
            If a ``Terminal`` passed to ``terminal`` does not belong to this component.
        AttributeError
            If a string passed to ``terminal`` does not represent an existing attribute.
        ValueError
            If a string passed to ``terminal`` does not represent an attribute of this
            component that produces a ``Terminal`` instance.

        See Also
        --------
        components.base.terminal.Terminal.reset_current()
        """
        terminal = self._get_or_check_terminal(terminal)
        terminal.reset_current(label, **kwargs)
        return self

    def clear_current(self, terminal: Terminal | str | None = None) -> Self:
        """Clear the current label on one of the terminals of the component.

        Parameters
        ----------
        terminal : Terminal | str | None
            Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a component (e.g. ``"right"``).
            If unspecified, defaults to the first terminal in the internal terminal
            list. In the case of monopoles, this is the only terminal, and in the case
            of bipoles, this is the left terminal.

        Raises
        ------
        ValueError
            If a ``Terminal`` passed to ``terminal`` does not belong to this component.
        AttributeError
            If a string passed to ``terminal`` does not represent an existing attribute.
        ValueError
            If a string passed to ``terminal`` does not represent an attribute of this
            component that produces a ``Terminal`` instance.

        See Also
        --------
        components.base.terminal.Terminal.clear_current()
        """
        terminal = self._get_or_check_terminal(terminal)
        terminal.clear_current()
        return self

    def voltage(
        self,
        start: Terminal | str,
        end: Terminal | str,
        *args: Any,
        **kwargs: Any,
    ) -> Voltage:
        """Return a voltage arrow across the component.

        Convenience method for creating a voltage arrow across two terminals of this
        component. Returns the created ``Voltage`` object. This method automatically
        sets the component is it called upon in the ``avoid`` argument of ``Voltage``
        (and as such overrides this argument).

        Parameters
        ----------
        start : Terminal | str
            Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a terminal (e.g. ``"right"``).
        end : Terminal | str
            Either a ``Terminal`` belonging to this component, or a string representing
            an attribute of this component that returns a terminal (e.g. ``"left"``).
        *args
            Positional arguments to be passed to the ``Voltage`` constructor.
        **kwargs
            Keyword arguments to be passed to the ``Voltage`` constructor. Any keyword
            argument with the key ``avoid`` will be ignored.

        Returns
        -------
        Voltage
            The voltage arrow resulting from the specification given.

        Raises
        ------
        ValueError
            If a passed ``Terminal`` does not belong to this component.
        AttributeError
            If a string passed for either terminal does not represent an existing
            attribute.
        ValueError
            If a string passed for either terminal does not represent an attribute of
            this component that produces a ``Terminal`` instance.
        ValueError
            If the terminals specified for both ``start`` and ``end`` are the same.
        """
        start = self._get_or_check_terminal(start)
        end = self._get_or_check_terminal(end)

        if start == end:
            raise ValueError(
                "The terminals specified through `start` and `end` are "
                "identical. They must be different."
            )

        kwargs["avoid"] = self

        return Voltage(start, end, *args, **kwargs)

    def _get_or_check_terminal(self, terminal: Terminal | str | None) -> Terminal:
        """Get a terminal or check a passed terminal belongs to this component.

        Parameters
        ----------
        terminal : Terminal | str | None
            The string to use as a terminal identifier, a ``Terminal`` instance to
            verify belongs to this component, or ``None``, in which case the first
            terminal will be used.

        Returns
        -------
        Terminal
            The terminal identified.

        Raises
        ------
        AttributeError
            If the string passed for the terminal doesn't exist as an attribute on this
            component.
        ValueError
            If the attribute identified by the string isn't an instance of ``Terminal``.
        ValueError
            If the terminal passed doesn't belong to this component.
        """
        if terminal is None:
            return self.terminals[0]

        if isinstance(terminal, Terminal):
            if terminal not in self.terminals:
                raise ValueError("Passed terminal does not belong to this component.")
            return terminal

        to_return = getattr(self, terminal)
        if not isinstance(to_return, Terminal):
            raise ValueError(
                f"Attribute `{terminal}` of `{self.__class__.__name__}` "
                f"is not a terminal."
            )
        return to_return

    def __set_up_anchors(self) -> None:
        # A small amount is added to each of these anchors to make sure that they are
        # never directly over the centre anchor, as this causes problems.
        self._label_anchor.shift(self._body.get_top() + 0.01 * mn.UP)
        self._annotation_anchor.shift(self._body.get_bottom() + 0.01 * mn.DOWN)
        self.add(self._centre_anchor, self._label_anchor, self._annotation_anchor)

    def __initialise_marks(
        self, label: str | Value | None, annotation: str | Value | None
    ) -> None:
        if label is not None:
            self.set_label(label)
        if annotation is not None:
            self.set_annotation(annotation)

    @mn.override_animate(set_label)
    def __animate_set_label(
        self, label: str | Value, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return self.animate(**anim_args)._set_mark(self._label, label).build()

    @mn.override_animate(clear_label)
    def __animate_clear_label(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return self.animate(**anim_args)._clear_mark(self._label).build()

    @mn.override_animate(set_annotation)
    def __animate_set_annotation(
        self, annotation: str | Value, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return self.animate(**anim_args)._set_mark(self._annotation, annotation).build()

    @mn.override_animate(clear_annotation)
    def __animate_clear_annotation(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return self.animate(**anim_args)._clear_mark(self._annotation).build()

    @mn.override_animate(set_current)
    def __animate_set_current(
        self,
        label: str,
        terminal: Terminal | str | None = None,
        anim_args: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        terminal = self._get_or_check_terminal(terminal)
        return terminal.animate(**anim_args).set_current(label, **kwargs).build()

    @mn.override_animate(reset_current)
    def __animate_reset_current(
        self,
        label: str,
        terminal: Terminal | str | None = None,
        anim_args: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        terminal = self._get_or_check_terminal(terminal)
        return terminal.animate(**anim_args).reset_current(label, **kwargs).build()

    @mn.override_animate(clear_current)
    def __animate_clear_current(
        self,
        terminal: Terminal | str | None = None,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        terminal = self._get_or_check_terminal(terminal)
        return terminal.animate(**anim_args).clear_current().build()



"""Base classes for creating sources."""

import abc
from typing import Any, Self

import manim as mn
import numpy as np

from manim_eng import config_eng
from manim_eng.components.base.bipole import SquareBipole
from manim_eng.components.base.terminal import Terminal

__all__ = [
    "CurrentSourceBase",
    "EuropeanCurrentSourceBase",
    "EuropeanVoltageSourceBase",
    "Source",
    "VoltageSourceBase",
]


class Source(SquareBipole, metaclass=abc.ABCMeta):
    """Base class of all sources."""

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            **kwargs,
        )

    @property
    def negative(self) -> Terminal:
        """Return the negative (left-hand) terminal of the source."""
        return self.left

    @property
    def positive(self) -> Terminal:
        """Return the positive (right-hand) terminal of the source."""
        return self.right

    @property
    def anode(self) -> Terminal:
        """Return the anode (positive terminal) of the source."""
        return self.positive

    @property
    def cathode(self) -> Terminal:
        """Return the cathode (negative terminal) of the source."""
        return self.negative


class VoltageSourceBase(Source, metaclass=abc.ABCMeta):
    """Base class of all voltage sources."""

    def __init__(self, arrow: bool, voltage: str | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.should_arrow = arrow
        self.__arrow: mn.Arrow | None = None

        if voltage is not None:
            self.set_voltage(voltage)

    def set_voltage(self, voltage: str) -> Self:
        """Set the voltage of the source.

        Sets the voltage label of the source, using the label of the component to do so.
        For non-arrowed (American-style) sources there is no difference between using
        this and using `.set_label()`, however for arrowed sources an arrow is added
        only when calling this method. For portability between source types it is
        therefore recommended to use this method over `.set_label()`.

        Parameters
        ----------
        voltage : str
            The voltage label to set. Takes a TeX math mode string.
        """
        if self.should_arrow and self.__arrow is None:
            self.__construct_arrow()
        self.set_label(voltage)
        return self

    def clear_voltage(self) -> Self:
        """Clear the voltage label of the source.

        Clears the voltage label of the source (i.e. the label of the component). For
        non-arrowed (American-style) sources there is no difference between using this
        and using `.clear_label()`, however for European sources the arrow is removed
        only when calling this method. For portability between source types it is
        therefore recommended to use this method over `.clear_label()`.
        """
        if self.should_arrow:
            self.__clear_arrow()
        self.clear_label()
        return self

    def __construct_arrow(self) -> None:
        half_side_length = config_eng.symbol.square_bipole_side_length / 2

        direction_to_label_anchor = mn.normalize(
            self._label_anchor.pos - self._centre_anchor.pos
        )
        arrow_midpoint_distance = 1.4 * half_side_length
        arrow_midpoint = (
            self._centre_anchor.pos
            + direction_to_label_anchor * arrow_midpoint_distance
        )
        arrow_half_length = 0.9 * half_side_length
        arrow_direction = np.cross(direction_to_label_anchor, mn.OUT)

        self.__arrow = mn.Arrow(
            start=arrow_midpoint - arrow_half_length * arrow_direction,
            end=arrow_midpoint + arrow_half_length * arrow_direction,
            tip_length=config_eng.symbol.arrow_tip_length,
            buff=0,
        ).match_style(self)
        self._label_anchor.move_to(arrow_midpoint)
        self._body.add(self.__arrow)

    def __clear_arrow(self) -> None:
        # It's unnecessary to move the label anchor to its original position, as the
        # next setting of the label will only put it back where it is now
        self._body.remove(self.__arrow)
        self.__arrow = None

    @mn.override_animate(set_voltage)
    def __animate_set_voltage(
        self, label: str, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        label_animation = self.animate(**anim_args).set_label(label).build()
        animations = [label_animation]

        if self.should_arrow and self.__arrow is None:
            self.__construct_arrow()
            arrow_animation = mn.Create(self.__arrow, **anim_args)
            animations.append(arrow_animation)

        return mn.AnimationGroup(*animations)

    @mn.override_animate(clear_voltage)
    def __animate_clear_voltage(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        label_animation = self.animate(**anim_args).clear_label().build()
        animations = [label_animation]

        if self.should_arrow and self.__arrow is not None:
            arrow_animation = mn.Uncreate(self.__arrow, **anim_args)
            animations.append(arrow_animation)
        self.__clear_arrow()

        return mn.AnimationGroup(*animations)


class EuropeanVoltageSourceBase(VoltageSourceBase, metaclass=abc.ABCMeta):
    """Base class for all European voltage sources.

    Designed to be used in conjunction with the ``RoundOuter`` and ``DiamondOuter``
    modifiers, to produce standard and controlled sources.

    Parameters
    ----------
    voltage : str | None
        The voltage label to set initially. Takes a TeX math mode string.

    See Also
    --------
    modifiers.RoundOuter
    modifiers.DiamondOuter
    """

    def __init__(self, voltage: str | None = None, **kwargs: Any) -> None:
        super().__init__(arrow=True, voltage=voltage, **kwargs)

    def _construct(self) -> None:
        super()._construct()
        half_width = config_eng.symbol.square_bipole_side_length / 2
        self._body.add(
            mn.Line(
                start=mn.LEFT * half_width,
                end=mn.RIGHT * half_width,
                z_index=self.z_index + 0.1,
            ).match_style(self)
        )


class CurrentSourceBase(Source, metaclass=abc.ABCMeta):
    """Base class of all current sources."""

    def __init__(self, current: str | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if current is not None:
            self.set_current(current)


class EuropeanCurrentSourceBase(CurrentSourceBase, metaclass=abc.ABCMeta):
    """Base class for all European current sources.

    Implements the setting of current on the positive terminal that is unique to
    European sources, as well as the internal symbol of European current sources.

    Designed to be used in conjunction with the ``RoundOuter`` and ``DiamondOuter``
    modifiers, to produce standard and controlled sources.

    See Also
    --------
    modifiers.RoundOuter
    modifiers.DiamondOuter
    """

    def _construct(self) -> None:
        super()._construct()
        half_width = config_eng.symbol.square_bipole_side_length / 2
        self._body.add(
            mn.Line(
                start=mn.UP * half_width,
                end=mn.DOWN * half_width,
                z_index=self.z_index + 0.1,
            ).match_style(self)
        )



"""Contains the Monopole base class."""

from __future__ import annotations

import abc
from typing import Any, Self

import manim as mn
import manim.typing as mnt

from manim_eng.components.base.component import Component
from manim_eng.components.base.terminal import Terminal
from manim_eng.components.node import Node
from manim_eng.units import Value

__all__ = ["Monopole"]


class Monopole(Component, metaclass=abc.ABCMeta):
    """Base class for monopole components, such as grounds and rails.

    Creates a single terminal in the direction of ``direction`` with its start at the
    origin.

    Parameters
    ----------
    direction : Vector3D
        The direction the terminal of the component should face.
    """

    def __init__(self, direction: mnt.Vector3D, **kwargs: Any) -> None:
        terminal = Terminal(
            position=mn.ORIGIN,
            direction=direction,
        )
        super().__init__(terminals=[terminal], **kwargs)

        self._label_anchor.move_to(self.get_critical_point(-direction))
        self.update()
        self.remove(self._annotation_anchor)

    @property
    def terminal(self) -> Terminal:
        """Get the terminal of the component."""
        return self.terminals[0]

    def align_monopole(
        self,
        other: Terminal | mnt.Point3D | Node | Monopole,
        direction: mnt.Vector3D | None = None,
    ) -> Self:
        """Aligns the monopole's terminal with another point or component.

        Moves this component along the line perpendicular to ``direction`` such that the
        line between the end of this component's terminal and ``other``
        has direction vector ``direction``.

        Parameters
        ----------
        other : Terminal | Point3D | Node | Monopole
            A ``Terminal`` belonging to another component, a ``Node``, a ``Monopole``
            (for which its single terminal is selected), or a point in space.
        direction : Vector3D | None
            The direction to align the terminals in. If not supplied, uses
            ``self_terminal``'s direction.

        Raises
        ------
        ValueError
            If ``other`` belongs to this component (if it is a ``Terminal``)
            or if ``other`` *is* this component (if it is a ``Node`` or
            ``Monopole``).

        Notes
        -----
        In geometric terms, the component in moved such that the end of
        this monopole's terminal is at the intersection of the lines that

        - Have direction vector perpendicular to ``direction`` and go through the
            current position of the end of this monopole's terminal; and
        - Have direction vector ``direction`` and go through the end of
            ``other`` (in the case that it is a ``Terminal``) or through ``other`` (in
            the case that it is a point).
        """
        return super().align_terminal(self.terminal, other, direction)

    def set_annotation(self, annotation: str | Value) -> Self:
        """Fails for monopoles, as they do not have annotations."""
        raise NotImplementedError(
            "Monopoles have no annotation. Please use `.set_label()`."
        )

    def clear_annotation(self) -> Self:
        """Fails for monopoles, as they do not have annotations."""
        raise NotImplementedError(
            "Monopoles have no annotation. Please use `.clear_label()`."
        )



"""Terminal base class and implementation helper class."""

from typing import Any, Self

import manim as mn
import manim.typing as mnt
import numpy as np
from manim import VMobject

from manim_eng._base.anchor import CentreAnchor, CurrentAnchor, TerminalAnchor
from manim_eng._base.mark import Mark
from manim_eng._base.markable import Markable
from manim_eng._config import config_eng
from manim_eng.units import Value

__all__ = ["Terminal"]


class CurrentArrow(mn.Triangle):
    def __init__(self, position: mnt.Vector3D, rotation: float = 0) -> None:
        super().__init__(
            radius=config_eng.symbol.current_arrow_radius,
            start_angle=0,
            color=mn.WHITE,
            fill_opacity=1.0,
        )
        self.move_to(position).rotate(rotation, about_point=position)


class Terminal(Markable):
    """Terminal for a circuit component (i.e. the bit other wires connect to).

    Parameters
    ----------
    position : Vector3D
        The position of the *start* of the terminal, i.e. the bit that 'connects' to the
        body of the component.
    direction : Vector3D
        The direction the terminal 'points', i.e. the direction you get by walking from
        the point on the component body where the terminal attaches to the end of the
        terminal.
    auto : bool
        Whether the terminal should control its visibility. When set to ``True``, the
        terminal will only be shown if there is at least one connection to it *or* there
        is a current annotation set on the terminal. When set to ``False``, the terminal
        is displayed no matter what. This is the default.
    """

    def __init__(
        self, position: mnt.Vector3D, direction: mnt.Vector3D, auto: bool = False
    ) -> None:
        super().__init__()

        self.autovisibility = auto
        self._connection_count = 0

        direction /= np.linalg.norm(direction)
        end = position + (direction * config_eng.symbol.terminal_length)
        self._line = mn.Line(
            start=position,
            end=end,
            stroke_width=config_eng.symbol.wire_stroke_width,
        )
        if not self.autovisibility:
            self.add(self._line)

        self._centre_anchor = CentreAnchor().move_to(self._line.get_center())
        self._end_anchor = TerminalAnchor().move_to(end)

        self._current_arrow: CurrentArrow
        self._current_arrow_showing: bool = False
        self._current_arrow_pointing_out: bool = False
        self.__rebuild_current_arrow()

        arrow_half_height = self._current_arrow.height / 2
        perpendicular = np.cross(direction, mn.IN)
        self._top_anchor = CurrentAnchor().move_to(
            self._centre_anchor.pos + arrow_half_height * perpendicular
        )
        self._bottom_anchor = CurrentAnchor().move_to(
            self._centre_anchor.pos - arrow_half_height * perpendicular
        )

        self.add(
            self._centre_anchor, self._end_anchor, self._top_anchor, self._bottom_anchor
        )

        self._current: Mark = Mark(self._top_anchor, self._centre_anchor)
        self._current_mark_anchored_below: bool = False

    @property
    def direction(self) -> mnt.Vector3D:
        """Return the direction of the terminal as a normalised vector."""
        return mn.normalize(self._end_anchor.pos - self._centre_anchor.pos)

    @property
    def end(self) -> mnt.Point3D:
        """Return the global position of the end of the terminal."""
        return self._end_anchor.pos

    def set_current(
        self,
        label: str | Value | None,
        out: bool | None = None,
        below: bool | None = None,
    ) -> Self:
        """Set the current label of the terminal.

        Sets the current label, with unspecified arguments being left to their current
        values.

        Parameters
        ----------
        label : str | Value, optional
            The current label to set. Takes a TeX math mode string, or a ``Value`` to be
            typeset as a math mode string.
        out : bool, optional
            Whether the arrow accompanying the annotation should point out (away from
            the body of the component to which the terminal is attached), or in (towards
            the component). If unspecified, falls back to the previous setting, or if
            there is none, the default (``False``).
        below : bool, optional
            Whether the annotation should be placed below the current arrow, or above
            it. Note that 'below' here is defined as below the terminal when it is
            pointing right. If unspecified, falls back to the previous setting, or if
            there is none, the default (``False``).

        See Also
        --------
        reset_current: Set the current label, with unspecified arguments being reset
                       to default.
        """
        if not self._current_arrow_showing:
            self.__rebuild_current_arrow()
            self.add(self._current_arrow)
            self._current_arrow_showing = True

        if out is not None and out != self._current_arrow_pointing_out:
            self._current_arrow.rotate(mn.PI, about_point=self._centre_anchor.pos)
            self._current_arrow_pointing_out = out

        if below is not None and below != self._current_mark_anchored_below:
            self._current.change_anchors(
                self._bottom_anchor if below else self._top_anchor,
                self._centre_anchor,
            )
            self._current_mark_anchored_below = below

        if label is not None:
            self._set_mark(self._current, label)

        self.__update_terminal_visibility()
        return self

    def reset_current(
        self, label: str | Value, out: bool = False, below: bool = False
    ) -> Self:
        """Set the current label of the terminal. Unspecified arguments are reset.

        Sets the current label, with unspecified arguments being reset to their original
        (default) values. In contrast to its sister method `.set_current()`, this method
        will always produce the same result regardless of where it is called.

        Parameters
        ----------
        label : str | Value
            The current label to set. Takes a TeX math mode string, or a ``Value`` to be
            typeset as a math mode string.
        out : bool
            Whether the arrow accompanying the annotation should point out (away from
            the body of the component to which the terminal is attached), or in (towards
            the component, this is the default).
        below : bool
            Whether the annotation should be placed below the current arrow, or above it
            (which is the default). Note that 'below' here is defined as below the
            terminal when it is pointing right.

        See Also
        --------
        set_current: Set the current label without resetting unspecified arguments.
        """
        return self.set_current(label=label, out=out, below=below)

    def clear_current(self) -> Self:
        """Clear the current annotation of the terminal."""
        self.remove(self._current_arrow)
        self._current_arrow_showing = False
        self._clear_mark(self._current)
        self.__update_terminal_visibility()
        return self

    def is_visible(self) -> bool:
        """Whether the terminal is currently visible on screen."""
        return (not self.autovisibility) or (
            self._connection_count > 0 or self._current_arrow_showing
        )

    def match_style(self, vmobject: VMobject, _family: bool = True) -> Self:
        """Match the style of the terminal wire to another vmobject.

        Parameters
        ----------
        vmobject : VMobject
            The vmobject to match to.
        _family : bool
            Disregarded in this case.

        Notes
        -----
        - It is not possible to override the stroke width
        - The ``_family`` argument has no effect.
        """
        self._line.match_style(vmobject).set_stroke(
            width=config_eng.symbol.wire_stroke_width
        )
        return self

    def _increment_connection_count(self) -> Self:
        self._connection_count += 1
        self.__update_terminal_visibility()
        return self

    def _decrement_connection_count(self) -> Self:
        self._connection_count -= 1
        self.__update_terminal_visibility()
        return self

    def __rebuild_current_arrow(self) -> None:
        """Rebuild the current arrow.

        Useful after an Uncreate or a rotation when the arrow wasn't in the scene (and
        therefore wasn't rotated).
        """
        angle_to_rotate = mn.angle_of_vector(self.direction)
        if not self._current_arrow_pointing_out:
            angle_to_rotate += np.pi
        self._current_arrow = CurrentArrow(self._centre_anchor.pos, angle_to_rotate)

    def __update_terminal_visibility(self) -> None:
        if not self.autovisibility:
            return

        if self._connection_count < 0:
            raise RuntimeError(
                f"Terminal cannot have negative connection count "
                f"({self._connection_count})."
            )

        should_be_visible = self._connection_count > 0 or self._current_arrow_showing
        if should_be_visible:
            self.add(self._line)
        else:
            self.remove(self._line)

    @mn.override_animate(set_current)
    def __animate_set_current(
        self,
        label: str | Value,
        out: bool | None = None,
        below: bool | None = None,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        animations: list[mn.Animation] = []

        rotation_needed = out is not None and out != self._current_arrow_pointing_out

        if not self._current_arrow_showing:
            self.__rebuild_current_arrow()
            self.add(self._current_arrow)
            self._current_arrow_showing = True
            if rotation_needed:
                self._current_arrow.rotate(mn.PI, about_point=self._centre_anchor.pos)
                self._current_arrow_pointing_out = out  # type: ignore[assignment]
            arrow_animation = mn.Create(
                self._current_arrow, introducer=False, **anim_args
            )
            animations.append(arrow_animation)

        elif rotation_needed:
            arrow_animation = mn.Rotate(
                self._current_arrow,
                mn.PI,
                about_point=self._centre_anchor.pos,
                **anim_args,
            )
            self._current_arrow_pointing_out = out  # type: ignore[assignment]
            animations.append(arrow_animation)

        if below is not None and below != self._current_mark_anchored_below:
            self._current.change_anchors(
                self._bottom_anchor if below else self._top_anchor,
                self._centre_anchor,
            )
            self._current_mark_anchored_below = below

        label_animation = (
            self.animate(**anim_args)._set_mark(self._current, label).build()
        )
        animations.append(label_animation)

        visibility_change_needed = self.autovisibility and self._connection_count == 0
        if visibility_change_needed:
            # The terminal is not yet showing, so 'create' it
            self.add(self._line)
            terminal_animation = mn.Create(self._line, introducer=False, **anim_args)
            animations.append(terminal_animation)

        return mn.AnimationGroup(*animations)

    @mn.override_animate(reset_current)
    def __animate_reset_current(
        self,
        label: str | Value,
        out: bool = False,
        below: bool = False,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}
        return (
            self.animate(**anim_args)
            .set_current(label=label, out=out, below=below)
            .build()
        )

    @mn.override_animate(clear_current)
    def __animate_clear_current(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation:
        if anim_args is None:
            anim_args = {}

        arrow_animation = mn.Uncreate(self._current_arrow, **anim_args)
        label_animation = self.animate(**anim_args)._clear_mark(self._current).build()
        animations: list[mn.Animation] = [arrow_animation, label_animation]

        if self.autovisibility and self._connection_count == 0:
            # The current arrow was the only reason for the terminal being shown, so
            # 'uncreate' it
            terminal_animation = mn.Uncreate(self._line, **anim_args)
            animations.append(terminal_animation)

        self._current_arrow_showing = False

        return mn.AnimationGroup(*animations)

    @mn.override_animate(_increment_connection_count)
    def __animate_increment_connection_count(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        self._connection_count += 1
        if not self.autovisibility:
            return None

        if anim_args is None:
            anim_args = {}

        terminal_already_visible = (
            self._connection_count > 1
        ) or self._current_arrow_showing
        self.__update_terminal_visibility()
        if terminal_already_visible:
            return None

        self.add(self._line)
        return mn.Create(self._line, introducer=False, **anim_args)

    @mn.override_animate(_decrement_connection_count)
    def __animate_decrement_connection_count(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        self._connection_count -= 1
        if not self.autovisibility:
            return None

        if anim_args is None:
            anim_args = {}

        self.__update_terminal_visibility()
        return mn.Uncreate(self._line, **anim_args)



"""Base classes for bipole components."""

import abc
from typing import Any

import manim as mn

from manim_eng import config_eng
from manim_eng.components.base.component import Component
from manim_eng.components.base.terminal import Terminal

__all__ = ["Bipole", "SquareBipole"]


class Bipole(Component, metaclass=abc.ABCMeta):
    """Base class for bipole components, such as resistors and sources.

    By default, adds two terminals: one from (-0.5, 0) to (-1, 0), and one from (0.5, 0)
    to (1, 0).

    Parameters
    ----------
    left : Terminal | None
        The terminal to use as the left connection point for the component. If left
        unspecified, the terminal will be in the default position for the left terminal
        of a rectangular bipole.
    right : Terminal | None
        The terminal to use as the right connection point for the component. If left
        unspecified, the terminal will be in the default position for the right terminal
        of a rectangular bipole.
    """

    def __init__(
        self,
        left: Terminal | None = None,
        right: Terminal | None = None,
        **kwargs: Any,
    ) -> None:
        half_width = config_eng.symbol.bipole_width / 2

        left = (
            left
            if left is not None
            else Terminal(position=half_width * mn.LEFT, direction=mn.LEFT)
        )
        right = (
            right
            if right is not None
            else Terminal(position=half_width * mn.RIGHT, direction=mn.RIGHT)
        )
        super().__init__(terminals=[left, right], **kwargs)

    def _construct(self) -> None:
        pass

    @property
    def left(self) -> Terminal:
        """Return the left-hand terminal of the component.

        Note that 'left' here is defined as when the component is unrotated. This does
        not adapt to rotation.
        """
        return self.terminals[0]

    @property
    def right(self) -> Terminal:
        """Return the right-hand terminal of the component.

        Note that 'right' here is defined as when the component is unrotated. This does
        not adapt to rotation.
        """
        return self.terminals[1]


class SquareBipole(Bipole, metaclass=abc.ABCMeta):
    """Base class for bipole components with a more square footprint.

    Parameters
    ----------
    left : Terminal | None
        The terminal to use as the left connection point for the component. If left
        unspecified, the terminal will be in the default position for the left terminal
        of a square bipole.
    right : Terminal | None
        The terminal to use as the right connection point for the component. If left
        unspecified, the terminal will be in the default position for the right terminal
        of a square bipole.
    """

    def __init__(
        self,
        left: Terminal | None = None,
        right: Terminal | None = None,
        **kwargs: Any,
    ) -> None:
        half_width = config_eng.symbol.square_bipole_side_length / 2
        super().__init__(
            Terminal(
                position=mn.LEFT * half_width,
                direction=mn.LEFT,
            )
            if left is None
            else left,
            Terminal(
                position=mn.RIGHT * half_width,
                direction=mn.RIGHT,
            )
            if right is None
            else right,
            **kwargs,
        )



"""Importer for classes useful in implementing custom components.

Import this file if you wish to define your own components. It brings in the base and
helper classes that manim-eng uses to build the public-facing component library.
"""

from ._base import *
from .circuits.base import *
from .components.base import *



"""Mandatory xkcd: https://xkcd.com/2818/.

The components (or aliases) of Randall Munroe's circuit symbols.
"""

from typing import Any

from manim_eng.components.base.xkcd import RandalMunroeSourceBase
from manim_eng.components.capacitors import Capacitor
from manim_eng.components.diodes import Photodiode
from manim_eng.components.monopoles import Earth
from manim_eng.components.switches import Switch

__all__ = [
    "Baertty",
    "Battttttttttttery",
    "CheckOutThisReallyCoolDiode",
    "Drawbridge",
    "Overpass",
    "PogoStick",
]


Drawbridge = Switch
Overpass = Capacitor
PogoStick = Earth
CheckOutThisReallyCoolDiode = Photodiode


class Baertty(RandalMunroeSourceBase):
    """Circuit symbol for Randall Munroe's baertty.

    See https://xkcd.com/2818/.

    Parameters
    ----------
    voltage : str | None
        Voltage label to set on creation, if desired. Takes a TeX math mode string.
    """

    def __init__(self, voltage: str | None = None, **kwargs: Any) -> None:
        super().__init__(
            pattern=[False, False, True, True],
            voltage=voltage,
            **kwargs,
        )


class Battttttttttttery(RandalMunroeSourceBase):
    """Circuit symbol for Randall Munroe's battttttttttttery.

    See https://xkcd.com/2818/.

    Parameters
    ----------
    voltage : str | None
        Voltage label to set on creation, if desired. Takes a TeX math mode string.
    """

    def __init__(self, voltage: str | None = None, **kwargs: Any) -> None:
        super().__init__(
            pattern=[False, True] + [False] * 6 + [True],
            voltage=voltage,
            **kwargs,
        )



"""Configuration classes and parser as well as manim-eng's default configuration."""

import dataclasses as dc
import re
from collections import defaultdict
from typing import Any, Self

import manim as mn
import numpy as np

STRING_TO_MANIM_COLOUR = {
    "white": mn.WHITE,
    "gray_a": mn.GRAY_A,
    "grey_a": mn.GREY_A,
    "gray_b": mn.GRAY_B,
    "grey_b": mn.GREY_B,
    "gray_c": mn.GRAY_C,
    "grey_c": mn.GREY_C,
    "gray_d": mn.GRAY_D,
    "grey_d": mn.GREY_D,
    "gray_e": mn.GRAY_E,
    "grey_e": mn.GREY_E,
    "black": mn.BLACK,
    "lighter_gray": mn.LIGHTER_GRAY,
    "lighter_grey": mn.LIGHTER_GREY,
    "light_gray": mn.LIGHT_GRAY,
    "light_grey": mn.LIGHT_GREY,
    "gray": mn.GRAY,
    "grey": mn.GREY,
    "dark_gray": mn.DARK_GRAY,
    "dark_grey": mn.DARK_GREY,
    "darker_gray": mn.DARKER_GRAY,
    "darker_grey": mn.DARKER_GREY,
    "blue_a": mn.BLUE_A,
    "blue_b": mn.BLUE_B,
    "blue_c": mn.BLUE_C,
    "blue_d": mn.BLUE_D,
    "blue_e": mn.BLUE_E,
    "pure_blue": mn.PURE_BLUE,
    "blue": mn.BLUE,
    "dark_blue": mn.DARK_BLUE,
    "teal_a": mn.TEAL_A,
    "teal_b": mn.TEAL_B,
    "teal_c": mn.TEAL_C,
    "teal_d": mn.TEAL_D,
    "teal_e": mn.TEAL_E,
    "teal": mn.TEAL,
    "green_a": mn.GREEN_A,
    "green_b": mn.GREEN_B,
    "green_c": mn.GREEN_C,
    "green_d": mn.GREEN_D,
    "green_e": mn.GREEN_E,
    "pure_green": mn.PURE_GREEN,
    "green": mn.GREEN,
    "yellow_a": mn.YELLOW_A,
    "yellow_b": mn.YELLOW_B,
    "yellow_c": mn.YELLOW_C,
    "yellow_d": mn.YELLOW_D,
    "yellow_e": mn.YELLOW_E,
    "yellow": mn.YELLOW,
    "gold_a": mn.GOLD_A,
    "gold_b": mn.GOLD_B,
    "gold_c": mn.GOLD_C,
    "gold_d": mn.GOLD_D,
    "gold_e": mn.GOLD_E,
    "gold": mn.GOLD,
    "red_a": mn.RED_A,
    "red_b": mn.RED_B,
    "red_c": mn.RED_C,
    "red_d": mn.RED_D,
    "red_e": mn.RED_E,
    "pure_red": mn.PURE_RED,
    "red": mn.RED,
    "maroon_a": mn.MAROON_A,
    "maroon_b": mn.MAROON_B,
    "maroon_c": mn.MAROON_C,
    "maroon_d": mn.MAROON_D,
    "maroon_e": mn.MAROON_E,
    "maroon": mn.MAROON,
    "purple_a": mn.PURPLE_A,
    "purple_b": mn.PURPLE_B,
    "purple_c": mn.PURPLE_C,
    "purple_d": mn.PURPLE_D,
    "purple_e": mn.PURPLE_E,
    "purple": mn.PURPLE,
    "pink": mn.PINK,
    "light_pink": mn.LIGHT_PINK,
    "orange": mn.ORANGE,
    "light_brown": mn.LIGHT_BROWN,
    "dark_brown": mn.DARK_BROWN,
    "gray_brown": mn.GRAY_BROWN,
    "grey_brown": mn.GREY_BROWN,
    "logo_white": mn.LOGO_WHITE,
    "logo_green": mn.LOGO_GREEN,
    "logo_blue": mn.LOGO_BLUE,
    "logo_red": mn.LOGO_RED,
    "logo_black": mn.LOGO_BLACK,
}


class ConfigBase:
    """Base class for manim-eng configuration classes."""

    def load_from_dict(
        self, dictionary: dict[str, Any], table_prefix: str = ""
    ) -> Self:
        r"""Load configuration in from a ``dict`` representation.

        Parameters
        ----------
        dictionary : dict[str, Any]
            The ``dict`` from which to load the values.
        table_prefix : str
            The current TOML table the dictionary values are a representation of. Allows
            this method to produce error messages that reflect the structure of the
            TOML from which the ``dict`` was generated.

        Notes
        -----
        This method is written as a strict intermediary between the configuration TOML
        file and the configuration classes. As such, an input of an empty dictionary
        ``{}`` will do nothing, as it is the equivalent of reading in an empty
        configuration file. The same goes for empty ``dict``\ s as values for tables: no
        change will be made to the table in this case.
        """
        possible_keys = self.__dict__.keys()

        for key, value in dictionary.items():
            if key not in possible_keys:
                raise ValueError(
                    f"Invalid {'table' if isinstance(value, dict) else 'key'} "
                    f"in manim-eng configuration: `{table_prefix}{key}`"
                )
            current_value = getattr(self, key)

            if isinstance(value, dict):
                # In this case, we have encountered a table.
                # First, check that this is a valid table.
                if not isinstance(current_value, ConfigBase):
                    raise ValueError(
                        f"Invalid table in manim-eng configuration: "
                        f"`{table_prefix}{key}`"
                    )
                # If it is, we call `load_from_dict` on the instance of ConfigBase that
                # represents the table.
                setattr(
                    self,
                    key,
                    type(current_value).load_from_dict(
                        current_value, value, table_prefix=f"{table_prefix}{key}."
                    ),
                )
                continue

            if isinstance(current_value, mn.ManimColor):
                if isinstance(value, mn.ManimColor):
                    setattr(self, key, value)
                    continue
                if not isinstance(value, str):
                    raise ValueError(
                        f"Invalid type in manim-eng configuration for key "
                        f"`{table_prefix}{key}`: "
                        f"{self._get_toml_type_from_python_variable(value)} "
                        f"(expected string)"
                    )
                if value not in STRING_TO_MANIM_COLOUR:
                    if not re.match(r"^#[0-9A-Fa-f]{6}$", value):
                        raise ValueError(
                            f"Invalid colour in manim-eng configuration "
                            f"for key `{table_prefix}{key}`: "
                            f"{value}"
                        )
                    setattr(self, key, mn.ManimColor(value))
                    continue
                setattr(self, key, STRING_TO_MANIM_COLOUR[value])
                continue

            if not isinstance(value, type(current_value)):
                raise ValueError(
                    f"Invalid type in manim-eng configuration for key "
                    f"`{table_prefix}{key}`: "
                    f"{self._get_toml_type_from_python_variable(value)} "
                    f"(expected "
                    f"{self._get_toml_type_from_python_variable(current_value)}"
                    f")"
                )

            setattr(self, key, value)

        return self

    def as_dict(self) -> dict[str, Any]:
        """Return this configuration as a dictionary.

        Returns
        -------
        dict[str, Any]
            This configuration as a dictionary, with subconfigurations being added as
            subdictionaries.
        """
        dictionary = self.__dict__
        for key, value in dictionary.items():
            if isinstance(value, ConfigBase):
                dictionary[key] = value.as_dict()
        return dictionary

    @staticmethod
    def _get_toml_type_from_python_variable(variable: Any) -> str:
        """Return the TOML type a Python variable would be stored as.

        Returns the TOML type that a Python value would've had if it were read from a
        manim-eng TOML configuration file. Serves to allow ``load_from_dict()`` to
        produce error messages relevant to the TOML the user wrote, rather than the
        internal Python representation of it.

        This is (roughly) an inversion of the table in the `tomllib docs <https://docs.python.org/3/library/tomllib.html#conversion-table>`_.
        """
        type_name = type(variable).__name__
        conversion_table = defaultdict(
            lambda: "table",
            {
                "str": "string",
                "int": "integer",
                "float": "float",
                "bool": "boolean",
                "list": "array",
                "dict": "table",
                "ManimColor": "string",
            },
        )
        return conversion_table[type_name]


@dc.dataclass
class ComponentSymbolConfig(ConfigBase):
    """Component display and behaviour configuration."""

    bipole_height: float = 0.4
    """The standard height to use for box-esque bipoles, such as resistors and fuses."""
    bipole_width: float = 1.0
    """The standard width to use for box-esque bipoles, such as resistors and fuses."""
    square_bipole_side_length: float = 1.5 * bipole_height
    """The standard height to use for bipoles with square bounding boxes, such as
    voltage sources and sensors."""
    component_stroke_width: float = mn.DEFAULT_STROKE_WIDTH
    """The stroke width to use for the component symbols."""
    current_arrow_radius: float = (2 / np.sqrt(3)) * 0.2 * bipole_height
    """The length from the centre of the current arrow triangle from its centre to one
    of its vertices."""
    terminal_length: float = 0.4 * bipole_width
    """The length of the terminal of a component."""
    wire_stroke_width: float = 0.625 * component_stroke_width
    """The stroke width to use for wires."""
    mark_font_size: float = 36.0
    """The default font size to use for marks (e.g. labels and annotations)."""
    mark_cardinal_alignment_margin: float = 5 * mn.DEGREES
    """The maximum angle a component can be from one of horizontal or vertical whilst
    still being considered horizontal or vertical for the purpose of mark alignment."""
    arrow_stroke_width: float = wire_stroke_width
    """The stroke width to use for arrows in voltage marks and similar."""
    arrow_tip_length: float = 0.2
    """The length of voltage arrow tips."""
    voltage_default_angle: float = 60 * mn.DEGREES
    """The angle a voltage arrow will sweep with no other reference provided."""
    node_radius: float = 0.06
    """The radius of wire nodes."""
    variability_arrow_tip_length: float = 0.125
    """The length of arrow tips in arrows signifying variability in a component."""
    monopole_width: float = 0.5 * bipole_width
    """The width of monopole source/ground symbols."""
    plate_gap: float = bipole_width / 6
    """The gap between plates of plated components (i.e. capacitors and cells)."""
    plate_height: float = 5 * plate_gap
    """The height of plates of plated components (i.e. capacitors and cells)."""


@dc.dataclass
class AnchorDisplayConfig(ConfigBase):
    """Anchor debug display configuration."""

    annotation_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.BLUE)
    """The colour to use for annotation anchors' debug visuals. Defaults to blue."""
    centre_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.PURPLE)
    """The colour to use for centre anchors' debug visuals. Defaults to purple."""
    current_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.ORANGE)
    """The colour to use for current anchors' debug visuals. Defaults to orange."""
    label_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.RED)
    """The colour to use for label anchors' debug visuals. Defaults to red."""
    radius: float = 0.06
    """The radius of anchor visualisation rings."""
    stroke_width: float = 2.0
    """The stroke width of anchor visualisation rings."""
    terminal_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.GREEN)
    """The colour to use for terminal anchors' debug visuals. Defaults to green."""
    voltage_colour: mn.ManimColor = dc.field(default_factory=lambda: mn.YELLOW)
    """The colour to use for voltage anchors' debug visuals. Defaults to yellow."""


@dc.dataclass
class ManimEngConfig(ConfigBase):
    """manim-eng configuration."""

    anchor: AnchorDisplayConfig = dc.field(default_factory=AnchorDisplayConfig)
    """Anchor debug display subconfig."""
    debug: bool = False
    """Whether or not to display debug information."""
    symbol: ComponentSymbolConfig = dc.field(default_factory=ComponentSymbolConfig)
    """Component symbol subconfig."""



"""Nodes for wire routing and display of circuit terminals and solder blobs."""

from typing import Any, Self, cast

import manim as mn
import manim.typing as mnt
import numpy as np

from manim_eng import config_eng
from manim_eng.components.base.component import Component
from manim_eng.components.base.terminal import Terminal
from manim_eng.units import Value

__all__ = ["Node", "OpenNode"]

AUTOBLOBBING_BLOB_THRESHOLD = 2
BLOB_Z_INDEX = 10


# Allows switch components to get an identical shape without bringing in all the extra
# baggage of the full Node class (anchors etc.)
def _create_node_blob(match_to: Component, open_: bool) -> mn.Dot:
    return mn.Dot(
        radius=config_eng.symbol.node_radius,
        stroke_width=config_eng.symbol.wire_stroke_width,
        stroke_color=match_to.color,
        fill_opacity=1.0,
        fill_color=mn.config.background_color if open_ else match_to.color,
        z_index=BLOB_Z_INDEX,
    )


class Node(Component):
    """Node in a circuit and open/filled terminal circuit symbol.

    ``Node`` handles two main purposes: it displays node symbols (open terminal symbols
    and solder blobs), and serves as an aid for wire routing, particularly when paired
    with updaters.

    Parameters
    ----------
    open_ : bool
        Whether to display an open or filled circle for the node. Open ones are
        typically used for external connections to a circuit (i.e. loose ends), whereas
        filled ones are used for 'solder blobs' to indicated that three or more wires
        connect.
    autoblob : bool
        Whether to handle the addition/removal of solder blobs automatically. Has no
        effect if the node is open (as autoblobbing only makes sense for solder blobs).
    """

    def __init__(self, open_: bool = False, autoblob: bool = True, **kwargs: Any):
        self.open = open_
        self.autoblob = autoblob if not open_ else False

        self.__blob: mn.Dot

        super().__init__(terminals=[], **kwargs)

        if self.autoblob:
            self.add_updater(self.__blob_updater)
            self.update()

        self.remove(self._annotation_anchor)

    def _construct(self) -> None:
        super()._construct()

        self.__blob = _create_node_blob(self, self.open)
        self._body.add(self.__blob)

    def set_label(
        self, label: str | Value, direction: mnt.Vector3D | float | None = None
    ) -> Self:
        """Set the label of the node, optionally specifying where it should be.

        The ``direction`` parameter can be used to specify the position the label should
        take. If it is left unspecified, the node will identify the most logical
        position for it (the widest gap between terminals) and keep the label here, with
        ties being broken by the uppermost position being favoured. This will continue
        to happen as terminals are added/removed.

        Parameters
        ----------
        label : str
            The label to set. Takes a TeX math mode string, or a ``Value`` to be typeset
            as a math mode string.
        direction : Vector3D | float | None
            The direction in which to place the label. Can either be a direction vector
            (``Vector3D``), an angle in radians (``float``), or ``None``, which
            signifies that the label should be placed automatically.
        """
        self._reposition_label_anchor(direction)
        super().set_label(label)
        return self

    def set_annotation(self, annotation: str | Value) -> Self:
        """Fails for nodes, as they do not have annotations."""
        raise NotImplementedError(
            "Nodes have no annotation. Please use `.set_label()`."
        )

    def clear_annotation(self) -> Self:
        """Fails for nodes, as they do not have annotations."""
        raise NotImplementedError(
            "Nodes have no annotation. Please use `.clear_label()`."
        )

    def get(self, direction: mnt.Vector3D | float) -> Terminal:
        """Get a terminal of the node in a given direction, creating it if necessary.

        Parameters
        ----------
        direction : mnt.Vector3D | float
            The direction to get a terminal in, as either a direction vector or an angle
            in radians. Note that the angle is defined as is mathematical standard:
            measured anticlockwise from the positive horizontal.

        Returns
        -------
        Terminal
            The terminal on the node in the specified direction.

        See Also
        --------
        clear
        right, up, left, down, up_right, up_left, down_left, down_right
        """
        direction = self._get_normalised_direction(direction)

        for terminal in self.terminals:
            if np.allclose(terminal.direction, direction):
                to_return = terminal
                break
        else:
            to_return = Terminal(
                position=self.get_center(),
                direction=direction,
                auto=True,
            ).match_style(self)
            self._terminals.add(to_return)

        return to_return

    @property
    def right(self) -> Terminal:
        """Get the right-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.RIGHT)

    @property
    def up(self) -> Terminal:
        """Get the up-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.UP)

    @property
    def left(self) -> Terminal:
        """Get the left-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.LEFT)

    @property
    def down(self) -> Terminal:
        """Get the down-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.DOWN)

    @property
    def up_right(self) -> Terminal:
        """Get the up-right-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.UR)

    @property
    def up_left(self) -> Terminal:
        """Get the up-left-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.UL)

    @property
    def down_left(self) -> Terminal:
        """Get the down-left-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.DL)

    @property
    def down_right(self) -> Terminal:
        """Get down-right-pointing terminal of the node, creating it if necessary."""
        return self.get(mn.DR)

    def set_blob_visibility(self, visible: bool) -> Self:
        """Alter the solder blob visibility.

        This will disable autoblobbing, as otherwise there would be two competing
        sources of truth on whether a blob should be displayed or not.

        Parameters
        ----------
        visible : bool
            Whether the solder blob should be visible.

        See Also
        --------
        show_blob
        hide_blob
        """
        self._set_blob_visibility(visible)
        self.disable_autoblobbing()
        return self

    def show_blob(self) -> Self:
        """Make the solder blob visible.

        This will disable autoblobbing, as otherwise there would be two competing
        sources of truth on whether a blob should be displayed or not.

        See Also
        --------
        set_blob_visibility
        hide_blob
        """
        return self.set_blob_visibility(visible=True)

    def hide_blob(self) -> Self:
        """Make the solder blob invisible.

        This will disable autoblobbing, as otherwise there would be two competing
        sources of truth on whether a blob should be displayed or not.

        See Also
        --------
        set_blob_visibility
        show_blob
        """
        return self.set_blob_visibility(visible=False)

    def set_autoblobbing(self, autoblob: bool) -> Self:
        """Specify whether the node should autoblob or not.

        If used to enable autoblobbing, an autoblob calculation will be made to decide
        whether to display the blob or not. Will not have an effect if the node is of an
        open type.

        Parameters
        ----------
        autoblob : bool
            Whether the node should autoblob or not.

        See Also
        --------
        enable_autoblobbing
        disable_autoblobbing
        """
        self.autoblob = autoblob
        if autoblob:
            if self.__blob_updater not in self.updaters:
                self.add_updater(self.__blob_updater)
            self.update()
        else:
            self.remove_updater(self.__blob_updater)
        return self

    def enable_autoblobbing(self) -> Self:
        """Enable autoblobbing for the node.

        If the node is of a filled type, an autoblob calculation will be made to decide
        whether to display the blob or not, and the node display updated accordingly.
        Will not have an effect if the node is of an open type.

        See Also
        --------
        set_autoblobbing
        disable_autoblobbing
        """
        return self.set_autoblobbing(True)

    def disable_autoblobbing(self) -> Self:
        """Disable autoblobbing for the node.

        Will not have an effect if the node is of an open type.

        See Also
        --------
        set_autoblobbing
        enable_autoblobbing
        """
        return self.set_autoblobbing(False)

    def make_open(self, make_visible: bool = True) -> Self:
        """Set the type of the node to open (an empty circle).

        Autoblobbing will be automatically disabled by this call. By default, it will
        also make the node symbol appear (i.e. an unfilled circle), regardless of
        whether it was showing before. Use the ``make_visible`` parameter to adjust this
        behaviour.

        Parameters
        ----------
        make_visible : bool
            Whether the open node symbol should be forced to become visible by this
            call. Note that a value of ``False`` will *not* force the node symbol to be
            invisible, but the symbol will maintain its previous visiblity. Defaults to
            ``True``.

        See Also
        --------
        make_filled
        """
        self.__blob.set_fill(color=mn.config.background_color)
        self.disable_autoblobbing()
        if make_visible:
            self.show_blob()
        return self

    def make_filled(self, reenable_autoblobbing: bool = True) -> Self:
        """Set the type of the node to filled (a filled solder blob, i.e. circle).

        By default, this call will automatically re-enable autoblobbing. To disable this
        behaviour, use the ``reenable_autoblobbing`` parameter.

        Parameters
        ----------
        reenable_autoblobbing : bool
            Whether to re-enable autoblobbing with this call. Defaults to ``True``.

        See Also
        --------
        make_open
        """
        self.__blob.set_fill(color=self.color)
        if reenable_autoblobbing:
            self.enable_autoblobbing()
        return self

    def get_center(self) -> mnt.Point3D:
        """Get the centre of the node.

        Note that this is not the geometric centre, but rather the point from which
        terminals originate (the centre of the node circle/blob).
        """
        return self.__blob.get_center()

    def _set_blob_visibility(self, visible: bool) -> Self:
        self.__blob.set_opacity(1.0 if visible else 0.0)
        return self

    def _get_normalised_direction(
        self, direction: mnt.Vector3D | float
    ) -> mnt.Vector3D:
        if isinstance(direction, float):
            return mn.rotate_vector(mn.RIGHT, direction)
        return mn.normalize(direction)

    def _should_be_visible(self) -> bool:
        visible_terminal_count = sum(
            [terminal.is_visible() for terminal in self.terminals]
        )
        return visible_terminal_count > AUTOBLOBBING_BLOB_THRESHOLD

    def _get_visible_terminal_angles(self) -> list[float]:
        return sorted(
            [
                mn.angle_of_vector(terminal.direction)
                for terminal in self.terminals
                if terminal.is_visible()
            ]
        )

    @staticmethod
    def _midangles_of_largest_gaps_between_list_of_angles(
        angles: list[float],
    ) -> list[float]:
        largest_gap = 0.0
        midangles: list[float] = []
        for start_angle, end_angle in zip(np.roll(angles, 1), angles, strict=False):
            if end_angle <= start_angle:
                # This comparison occurs over the 'break' at ±pi, or there is only one
                # element in `angles`. Either way, we get the behaviour we want by
                # adding 2 pi
                end_angle += 2 * np.pi  # noqa: PLW2901

            gap = end_angle - start_angle
            if gap < largest_gap and not np.isclose(gap, largest_gap):
                continue

            if gap > largest_gap:
                largest_gap = gap
                midangles = []

            centre_angle = 0.5 * (start_angle + end_angle)
            if centre_angle > np.pi:
                centre_angle -= 2 * np.pi
            midangles.append(centre_angle)

        return sorted(midangles)

    @staticmethod
    def _topmost_angle_as_direction(angles: list[float]) -> mnt.Vector3D:
        if len(angles) == 0:
            return mn.UP

        angles.sort(key=lambda x: np.sin(x), reverse=True)
        angle = angles[0]
        return mn.rotate_vector(mn.RIGHT, angle)

    def _get_optimal_label_anchor_direction(self) -> mnt.Vector3D:
        terminal_angles = self._get_visible_terminal_angles()
        angles = self._midangles_of_largest_gaps_between_list_of_angles(terminal_angles)
        return self._topmost_angle_as_direction(angles)

    def _update_label_positioning_using_vector(self, direction: mnt.Vector3D) -> None:
        position = self.get_center() + config_eng.symbol.node_radius * direction
        self._label_anchor.move_to(position)
        self._label.update()

    def _reposition_label_anchor(self, direction: mnt.Vector3D | float | None) -> None:
        if direction is None:
            if self.__label_anchor_updater not in self.updaters:
                self.add_updater(self.__label_anchor_updater)
            self.update()
        else:
            if self.__label_anchor_updater in self.updaters:
                self.remove_updater(self.__label_anchor_updater)
            direction = self._get_normalised_direction(direction)
            self._update_label_positioning_using_vector(direction)

    @staticmethod
    def __blob_updater(mobject: mn.Mobject) -> None:
        node = cast(Node, mobject)
        node._set_blob_visibility(node._should_be_visible())

    @staticmethod
    def __label_anchor_updater(mobject: mn.Mobject) -> None:
        node = cast(Node, mobject)
        new_direction = node._get_optimal_label_anchor_direction()
        node._update_label_positioning_using_vector(new_direction)

    @mn.override_animate(set_label)
    def __animate_set_label(
        self,
        label: str,
        direction: mnt.Vector3D | float | None = None,
        anim_args: dict[str, Any] | None = None,
    ) -> mn.Animation:
        self._reposition_label_anchor(direction)
        return self.animate(**anim_args)._set_mark(self._label, label).build()


class OpenNode(Node):
    """Open node circuit symbol.

    A utility wrapper around the ``Node`` class that sets the ``open_`` parameter to
    ``True`` automatically.

    See Also
    --------
    Node
    """

    def __init__(self, **kwargs: Any):
        super().__init__(open_=True, **kwargs)



"""Contains the Mark class.

See Also
--------
anchor, markable
"""

from typing import Any, Callable, Self

import manim as mn

from manim_eng._base.anchor import Anchor
from manim_eng._config import config_eng
from manim_eng._utils import utils

__all__ = ["Mark"]


class Mark(mn.VMobject):
    """A mark object, representing any textual annotation on a component.

    Parameters
    ----------
    anchor : Anchor
        The anchor to use as a base for the attachment of the mark.
    centre_reference : Anchor
        The anchor to use as a reference; the mark will be kept aligned to ``anchor``,
        attached to the side directly opposite the side ``centre_reference`` is on.

    See Also
    --------
    anchor.Anchor
    """

    def __init__(self, anchor: Anchor, centre_reference: Anchor) -> None:
        super().__init__()
        self.mathtex: mn.MathTex = mn.MathTex("")

        self.updater: Callable[[mn.Mobject], None]
        self.change_anchors(anchor, centre_reference)

    def set_text(
        self,
        *args: Any,
        font_size: float = config_eng.symbol.mark_font_size,
        **kwargs: Any,
    ) -> Self:
        """Set the text of the mark.

        Parameters
        ----------
        *args : Any
            Positional arguments to be pass on to ``manim.MathTex``. The most important
            of these is ``*tex_strings``, i.e. the actual TeX math mode strings to use
            as the mark's text.
        font_size : float
            The font size to use for the mark. Leaving it empty adopts the default
            (recommended).
        **kwargs : Any
            Keyword arguments to pass on to ``manim.MathTex``.
        """
        if self.mathtex in self.submobjects:
            self.remove(self.mathtex)
        self.mathtex = mn.MathTex(*args, font_size=font_size, **kwargs)
        self.add(self.mathtex)
        self._reposition()
        return self

    def change_anchors(self, anchor: Anchor, centre_reference: Anchor) -> None:
        """Change the anchors to which the mark is attached.

        Parameters
        ----------
        anchor : Anchor
            The anchor to use as a base for the attachment of the mark.
        centre_reference : Anchor
            The anchor to use as a reference; the mark will be kept aligned to
            ``anchor``, attached to the side directly opposite the side
            ``centre_reference`` is on.
        """
        if len(self.updaters) != 0:
            self.remove_updater(self.updater)
        self.updater = self.__get_updater(anchor, centre_reference)
        self.add_updater(self.updater)
        self.update()

    @property
    def tex_strings(self) -> list[str] | None:
        if self.mathtex is None:
            return None
        return self.mathtex.tex_strings  # type: ignore[no-any-return]

    def _reposition(self) -> None:
        """Update the mark position so it is realigned with its anchor."""
        # If updating is suspended, the mark cannot be repositioned, as positioning is
        # done through an updater. Hence, allow updating and perform an update, but then
        # suspend updating again, all non-recursively to prevent any other
        # 'non-essential' updates
        if self.updating_suspended:
            self.resume_updating(recursive=False)
            self.update(recursive=False)
            self.suspend_updating(recursive=False)
        else:
            self.update()

    @staticmethod
    def __get_updater(
        anchor: Anchor, centre_reference: Anchor
    ) -> Callable[[mn.Mobject], None]:
        if (anchor.pos == centre_reference.pos).all():
            raise ValueError(
                "`anchor` and `centre_reference` cannot be the same. "
                f"Found: {anchor.pos=}, {centre_reference.pos=}.\n"
                "Please report this error to a developer."
            )

        def updater(mark: mn.Mobject) -> None:
            line_of_connection = anchor.pos - centre_reference.pos
            line_of_connection = mn.normalize(line_of_connection)
            line_of_connection = utils.cardinalised(
                line_of_connection, config_eng.symbol.mark_cardinal_alignment_margin
            )
            mark.next_to(
                mobject_or_point=anchor.pos,
                direction=line_of_connection,
                buff=mn.SMALL_BUFF,
            )

        return updater



"""Component symbols of switches (both lever-arm and push-button)."""

from typing import Any, Self

import manim as mn

from manim_eng import config_eng
from manim_eng.components.base.switch import BipoleSwitchBase, PushSwitchBase

__all__ = ["PushToBreakSwitch", "PushToMakeSwitch", "Switch"]


class Switch(BipoleSwitchBase):
    """Circuit symbol for a basic two-terminal lever-arm switch.

    Parameters
    ----------
    closed : bool
        Whether the switch should be initially closed or not.
    """

    __open_wiper_angle: float = 30 * mn.DEGREES

    def __init__(self, closed: bool = False, **kwargs: Any) -> None:
        self.wiper: mn.Line

        super().__init__(closed, **kwargs)

    def _construct(self) -> None:
        super()._construct()

        self.wiper = (
            mn.Line(
                start=self.left_node.get_center(),
                end=self.right_node.get_center(),
                stroke_width=config_eng.symbol.component_stroke_width,
            )
            .match_style(self)
            .rotate(self.__open_wiper_angle, about_point=self.left_node.get_center())
        )

        self._body.add(self.wiper)

    def open(self) -> Self:
        """Open the switch, if not already open."""
        # For some reason Mypy can't work out the type of self.closed
        if not self.closed:  # type: ignore[has-type]
            return self
        self.wiper.rotate(
            self.__open_wiper_angle, about_point=self.left_node.get_center()
        )
        self.closed = False
        return self

    def close(self) -> Self:
        """Close the switch, if not already closed."""
        if self.closed:
            return self
        self.wiper.rotate(
            -self.__open_wiper_angle, about_point=self.left_node.get_center()
        )
        self.closed = True
        return self

    @mn.override_animate(open)
    def __animate_open(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        if anim_args is None:
            anim_args = {}
        if not self.closed:
            return None
        self.closed = False
        return mn.Rotate(
            self.wiper,
            angle=self.__open_wiper_angle,
            about_point=self.left_node.get_center(),
            **anim_args,
        )

    @mn.override_animate(close)
    def __animate_close(
        self, anim_args: dict[str, Any] | None = None
    ) -> mn.Animation | None:
        if anim_args is None:
            anim_args = {}
        if self.closed:
            return None
        self.closed = True
        return mn.Rotate(
            self.wiper,
            angle=-self.__open_wiper_angle,
            about_point=self.left_node.get_center(),
            **anim_args,
        )


class PushToMakeSwitch(PushSwitchBase):
    """Component symbol for a push-to-make switch.

    Parameters
    ----------
    closed : bool
        Whether the switch should be initially closed or not. Defaults to open.
    """

    def __init__(self, closed: bool = False, **kwargs: Any) -> None:
        super().__init__(push_to_make=True, closed=closed, **kwargs)


class PushToBreakSwitch(PushSwitchBase):
    """Component symbol for a push-to-break switch.

    Parameters
    ----------
    closed : bool
        Whether the switch should be initially closed or not. Defaults to **closed**
    (this is in contrast to the other switches, which default to open).
    """

    def __init__(self, closed: bool = True, **kwargs: Any) -> None:
        super().__init__(push_to_make=False, closed=closed, **kwargs)



"""Configuration management for manim-eng.

Exposes the configuration of manim-eng through the ``config_eng`` variable as well as
the temporary configuration context manager ``tempconfig_eng``.
"""

import contextlib
import copy
from typing import Any, Generator

from .config import ManimEngConfig
from .config_readers import get_project_config, get_user_config

__all__ = ["config_eng", "tempconfig_eng"]


config_eng = (
    ManimEngConfig()
    .load_from_dict(get_user_config())
    .load_from_dict(get_project_config())
)


@contextlib.contextmanager
def tempconfig_eng(temp_config: dict[str, Any]) -> Generator:
    """Context manager for temporary configuration.

    .. note::

        This does not change Manim's configuration, only that of manim-eng.

    Parameters
    ----------
    temp_config: dict[str, Any]
        Key-value pairs to use to temporary update the manim-eng configuration.

    Examples
    --------
    >>> config_eng.debug
    False
    >>> with tempconfig_eng({"debug": True}):
    ...     print(config_eng.debug)
    ...
    True
    >>> config_eng.debug
    False
    """
    global config_eng  # noqa: PLW0602
    original = copy.deepcopy(config_eng).as_dict()

    config_eng.load_from_dict(temp_config)

    try:
        yield
    finally:
        # Note that we load here instead of assigning so that we update the original
        # object.
        config_eng.load_from_dict(original)



"""Modifiers that can be applied to other components to add elements.

They are applied by inheritance, and should be the *first* class a subclass inherits
from. The other requirement is that the subclass implement ``._construct()`` and call
``super()._construct()``. A bare-minimum example is shown below.

::

    class ModifiedComponent(Modifier, ComponentToModify):
        def _construct(self) -> None:
            super()._construct()
"""

import abc

import manim as mn
import numpy as np

from manim_eng import config_eng
from manim_eng.components.base.component import Component

__all__ = ["DiamondOuter", "RoundOuter", "SensorModifier", "VariableModifier"]


class RoundOuter(Component, metaclass=abc.ABCMeta):
    """Modifier to add a circular outline to a component.

    The diameter of the outline is equal to the side length of a square bipole.
    """

    def _construct(self) -> None:
        super()._construct()
        self._body.add(
            # Use an Arc instead of a Circle because it doesn't have a strange default
            # colour
            mn.Arc(
                radius=config_eng.symbol.square_bipole_side_length / 2,
                angle=mn.TAU,
            ).match_style(self)
        )


class DiamondOuter(Component, metaclass=abc.ABCMeta):
    """Modifier to add a diamond outline to a component.

    The distance between opposite points of the outline is equal to the side length of a
    square bipole.
    """

    def _construct(self) -> None:
        super()._construct()
        self._body.add(
            mn.Square(
                side_length=config_eng.symbol.square_bipole_side_length / np.sqrt(2),
            )
            .match_style(self)
            .rotate(45 * mn.DEGREES)
        )


class VariableModifier(Component, metaclass=abc.ABCMeta):
    """Modifier to add diagonal arrow to a component to signify variability."""

    def _construct(self) -> None:
        super()._construct()

        component_width = float(
            np.linalg.norm(self._body.get_right() - self._body.get_left())
        )
        component_height = float(
            np.linalg.norm(self._body.get_bottom() - self._body.get_top())
        )

        arrow_half_height = component_height * (
            1.0 if component_width > component_height else 0.5
        )
        arrow_half_height = max(
            arrow_half_height, 0.5 * config_eng.symbol.square_bipole_side_length
        )
        arrow_half_width = 0.8 * arrow_half_height

        arrow = mn.Arrow(
            start=np.array([-arrow_half_width, -arrow_half_height, 0])
            + self._body.get_center(),
            end=np.array([arrow_half_width, arrow_half_height, 0])
            + self._body.get_center(),
            buff=0,
            tip_length=config_eng.symbol.variability_arrow_tip_length,
            stroke_width=self.stroke_width,
            stroke_color=self.stroke_color,
            stroke_opacity=self.stroke_opacity,
            fill_color=self.stroke_color,
            fill_opacity=self.stroke_opacity,
        )
        self._body.add(arrow)


class SensorModifier(Component, metaclass=abc.ABCMeta):
    """Modifier to add a 'bent L' to a component to signify that it is a sensor.

    The 'bent L' will be positioned such that the main arm will pass through the
    component symbol's centre.
    """

    __min_ratio = 0.8
    __max_ratio = 1.8

    def _construct(self) -> None:
        super()._construct()

        component_width = float(
            np.linalg.norm(self._body.get_right() - self._body.get_left())
        )
        component_height = float(
            np.linalg.norm(self._body.get_bottom() - self._body.get_top())
        )

        margin = config_eng.symbol.square_bipole_side_length * 0.2
        half_width = component_width / 2
        half_height = component_height / 2

        if component_width > component_height:
            half_height += margin
        else:
            half_width += margin

        ratio = half_width / half_height
        if ratio < self.__min_ratio:
            half_width = self.__min_ratio * half_height
        if ratio > self.__max_ratio:
            half_height = half_width / self.__max_ratio

        base_length = 0.5 * half_width

        bottom_left = np.array([-half_width, -half_height, 0])
        bottom_middle = np.array([-(half_width - base_length), -half_height, 0])
        top_right = np.array([half_width, half_height, 0])

        main_midpoint_offset = (bottom_middle + top_right) / 2

        tick = (
            mn.VMobject()
            .match_style(self)
            .set_points_as_corners([bottom_left, bottom_middle, top_right])
            .move_to(self._body.get_center() - main_midpoint_offset)
        )

        self._body.add(tick)



"""Wire and related implementation classes."""

from typing import Self, Sequence, cast

import manim as mn
import manim.typing as mnt
import numpy as np

from manim_eng._utils import utils
from manim_eng.circuits.base.wire import WireBase
from manim_eng.components.base.terminal import Terminal

__all__ = ["ManualWire", "Wire"]


class ManualWire(WireBase):
    """Wire that requires its path to be manually specified.

    Parameters
    ----------
    start : Terminal
        The terminal the wire starts at.
    end : Terminal
        The terminal the wire ends at.
    corner_points : Sequence[Point3D], optional
        The vertices the wire should have between the two terminals. Should not include
        the positions of the two terminals, as these are inserted automatically when the
        wire is drawn. These should be in order from ``start`` to ``end``. If left
        unspecified, this is taken to be ``[]`` and the wire will directly connect the
        start and end terminals.
    updating : bool
        Whether the ends of the wire should update automatically to keep connected to
        the terminals. This is disabled by default. If this is enabled, it is
        recommended to attach another updater that will update ``corner_points`` to
        prevent strange artefacts.

    Raises
    ------
    ValueError
        If ``start`` and ``end`` are the same.
    """

    def __init__(
        self,
        start: Terminal,
        end: Terminal,
        corner_points: Sequence[mnt.Point3D] | None = None,
        updating: bool = False,
    ):
        self._corner_points = list(corner_points) if corner_points is not None else []
        super().__init__(start, end, updating)

    def get_corner_points(self) -> list[mnt.Point3D]:
        """Get the corner points of the wire.

        Returns the vertices of the wire, not including the end points (i.e. at the
        start and end terminals).
        """
        return self._corner_points

    def set_corner_points(self, points: Sequence[mnt.Point3D]) -> Self:
        """Set the corner points of the wire.

        Parameters
        ----------
        points : Sequence[Point3D]
            The vertices the wire should have between the two terminals. Should not
            include the positions of the two terminals, as these are inserted
            automatically when the wire is drawn. These should be in order from
            ``start`` to ``end``.
        """
        self._corner_points = list(points)
        return self


class Wire(WireBase):
    """Wire to automatically connect components together.

    The connection algorithm will do its best to avoid going 'backwards' through
    components' terminals whilst ensuring that automatic connections have no more than
    two vertices and are only horizontal and vertical.

    Parameters
    ----------
    start : Terminal
        The terminal the wire starts at.
    end : Terminal
        The terminal the wire ends at.

    Raises
    ------
    ValueError
        If ``start`` and ``end`` are the same.
    """

    def __init__(self, start: Terminal, end: Terminal) -> None:
        super().__init__(start, end, updating=True)

    def get_corner_points(self) -> list[mnt.Point3D]:
        """Get the corner points of the wire.

        Returns the vertices of the wire, not including the end points (i.e. at the
        start and end terminals).
        """
        from_direction = utils.cardinalised(self.start.direction)
        to_direction = utils.cardinalised(self.end.direction)

        if np.isclose(np.dot(from_direction, to_direction), 0):
            return self.__get_corner_points_for_perpendicular_terminals(
                from_direction, to_direction
            )
        return self.__get_corner_points_for_parallel_terminals(
            from_direction, to_direction
        )

    def __get_corner_points_for_perpendicular_terminals(
        self, from_direction: mnt.Vector3D, to_direction: mnt.Vector3D
    ) -> list[mnt.Point3D]:
        from_end = self.start.end
        to_end = self.end.end

        corner_point = mn.find_intersection(
            [from_end], [from_direction], [to_end], [to_direction]
        )[0]

        if self.__point_is_behind_plane(
            corner_point, from_end, from_direction
        ) or self.__point_is_behind_plane(corner_point, to_end, to_direction):
            # Move the corner point to the other vertex of the box formed from the end
            # of each terminal, as two 90 degree turns at a component is better than one
            # 0 degree and one 180 degree.
            if corner_point[0] == from_end[0]:
                corner_point = np.array([to_end[0], from_end[1], 0])
            else:
                corner_point = np.array([from_end[0], to_end[1], 0])

        return [corner_point]

    def __get_corner_points_for_parallel_terminals(
        self, from_direction: mnt.Vector3D, to_direction: mnt.Vector3D
    ) -> list[mnt.Point3D]:
        midpoint = mn.midpoint(self.start.end, self.end.end)

        to_behind_from = self.__point_is_behind_plane(
            self.end.end, self.start.end, from_direction
        )
        from_behind_to = self.__point_is_behind_plane(
            self.start.end, self.end.end, to_direction
        )

        if to_behind_from and from_behind_to:
            # This is necessary to prevent the line from going backwards through the
            # components
            from_direction = mn.rotate_vector(from_direction, np.pi / 2)
            to_direction = mn.rotate_vector(to_direction, np.pi / 2)
        # These two are to handle the case where two terminals point in the same
        # direction, so we really want an elbow rather than an 'S'
        elif to_behind_from:
            midpoint = self.__move_point_forward_of_plane(
                midpoint, self.start.end, from_direction
            )
        elif from_behind_to:
            midpoint = self.__move_point_forward_of_plane(
                midpoint, self.end.end, to_direction
            )

        perpendicular_direction = np.cross(from_direction, mn.OUT)
        corner_points = mn.find_intersection(
            [midpoint] * 2,
            [perpendicular_direction] * 2,
            [self.start.end, self.end.end],
            [from_direction, to_direction],
        )
        return list(corner_points)

    @staticmethod
    def __point_is_behind_plane(
        point: mnt.Point3D, point_on_plane: mnt.Point3D, normal: mnt.Vector3D
    ) -> bool:
        """Return whether a given point is behind a specified plane.

        Parameters
        ----------
        point : Point3D
            The point to check.
        point_on_plane : Point3D
            A point on the plane against which to check.
        normal : Vector3D
            The normal vector of the plane against which to check.

        Returns
        -------
        bool
            ``True`` if the point is behind the plane, ``False`` if it is not.
        """
        vector_to_point = point - point_on_plane
        return cast(bool, np.dot(normal, vector_to_point) < 0)

    @staticmethod
    def __move_point_forward_of_plane(
        point: mnt.Point3D, point_on_plane: mnt.Point3D, normal: mnt.Vector3D
    ) -> mnt.Point3D:
        """Move a given point such that it lies on or in front of a specified plane.

        Parameters
        ----------
        point : Point3D
            The point to move.
        point_on_plane : Point3D
            A point on the plane.
        normal : Vector3D
            The normal vector of the plane

        Returns
        -------
        Point3D
            The new plane.
        """
        vector_to_point = point - point_on_plane
        distance_to_move = -np.dot(normal, vector_to_point)
        if distance_to_move <= 0:
            # No movement is necessary
            return point
        return point + mn.normalize(normal) * distance_to_move



"""Circuit symbols of all available components."""

from .capacitors import *
from .cells import *
from .diodes import *
from .inductors import *
from .monopoles import *
from .node import *
from .resistors import *
from .sources import *
from .switches import *
from .xkcd import *
